#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import dataclasses
from itertools import chain, product
from typing import Any, cast, Dict
from unittest import mock

import numpy as np
import torch
from ax.core.search_space import SearchSpaceDigest
from ax.exceptions.core import DataRequiredError
from ax.models.torch.botorch import (
    BotorchModel,
    get_feature_importances_from_botorch_model,
    get_rounding_func,
)
from ax.models.torch.botorch_defaults import (
    get_and_fit_model,
    get_chebyshev_scalarization,
    recommend_best_out_of_sample_point,
)
from ax.models.torch.utils import sample_simplex
from ax.models.torch_base import TorchOptConfig
from ax.utils.common.testutils import TestCase
from ax.utils.common.typeutils import not_none
from ax.utils.testing.mock import fast_botorch_optimize
from ax.utils.testing.torch_stubs import get_torch_test_data
from botorch.acquisition.utils import get_infeasible_cost
from botorch.models import ModelListGP, SingleTaskGP
from botorch.models.transforms.input import Warp
from botorch.utils.datasets import SupervisedDataset
from botorch.utils.objective import get_objective_weights_transform
from gpytorch.likelihoods import _GaussianLikelihoodBase
from gpytorch.likelihoods.gaussian_likelihood import FixedNoiseGaussianLikelihood
from gpytorch.mlls import ExactMarginalLogLikelihood, LeaveOneOutPseudoLikelihood
from gpytorch.priors import GammaPrior
from gpytorch.priors.lkj_prior import LKJCovariancePrior


FIT_MODEL_MO_PATH = f"{get_and_fit_model.__module__}.fit_gpytorch_mll"
SAMPLE_SIMPLEX_UTIL_PATH = f"{sample_simplex.__module__}.sample_simplex"
SAMPLE_HYPERSPHERE_UTIL_PATH = f"{sample_simplex.__module__}.sample_hypersphere"
CHEBYSHEV_SCALARIZATION_PATH = (
    f"{get_chebyshev_scalarization.__module__}.get_chebyshev_scalarization"
)


def dummy_func(X: torch.Tensor) -> torch.Tensor:
    return X


class BotorchModelTest(TestCase):
    def test_fixed_rank_BotorchModel(
        self, dtype: torch.dtype = torch.float, cuda: bool = False
    ) -> None:
        Xs1, Ys1, Yvars1, bounds, _, feature_names, metric_names = get_torch_test_data(
            dtype=dtype, cuda=cuda, constant_noise=True
        )
        Xs2, Ys2, Yvars2, _, _, _, _ = get_torch_test_data(
            dtype=dtype, cuda=cuda, constant_noise=True
        )
        model = BotorchModel(multitask_gp_ranks={"y": 2, "w": 1})
        datasets = [
            SupervisedDataset(
                X=Xs1[0],
                Y=Ys1[0],
                Yvar=Yvars1[0],
                feature_names=feature_names,
                outcome_names=["y"],
            ),
            SupervisedDataset(
                X=Xs2[0],
                Y=Ys2[0],
                Yvar=Yvars2[0],
                feature_names=feature_names,
                outcome_names=["w"],
            ),
        ]
        with self.assertRaisesRegex(RuntimeError, "Please fit the model first"):
            model.model

        with self.assertRaisesRegex(RuntimeError, "Please fit the model first"):
            model.search_space_digest

        search_space_digest = SearchSpaceDigest(
            feature_names=feature_names,
            bounds=bounds,
            task_features=[0],
        )
        with self.assertRaisesRegex(RuntimeError, "manually is disallowed"):
            model.search_space_digest = search_space_digest

        with mock.patch(FIT_MODEL_MO_PATH) as _mock_fit_model:
            model.fit(
                datasets=datasets,
                search_space_digest=search_space_digest,
            )
            self.assertTrue(isinstance(model.search_space_digest, SearchSpaceDigest))
            self.assertEqual(model.search_space_digest, search_space_digest)
            _mock_fit_model.assert_called_once()

        model.model = model.model  # property assignment isn't blocked
        # Check ranks
        model_list = cast(ModelListGP, model.model).models
        self.assertEqual(model_list[0]._rank, 2)
        self.assertEqual(model_list[1]._rank, 1)

    def test_fixed_prior_BotorchModel(
        self, dtype: torch.dtype = torch.float, cuda: bool = False
    ) -> None:
        Xs1, Ys1, Yvars1, bounds, _, feature_names, metric_names = get_torch_test_data(
            dtype=dtype, cuda=cuda, constant_noise=True
        )
        Xs2, Ys2, Yvars2, _, _, _, _ = get_torch_test_data(
            dtype=dtype, cuda=cuda, constant_noise=True
        )
        kwargs: Dict[str, Any] = {
            "prior": {
                "covar_module_prior": {
                    "lengthscale_prior": GammaPrior(6.0, 3.0),
                    "outputscale_prior": GammaPrior(3.0, 12.0),
                },
                "type": LKJCovariancePrior,
                "sd_prior": GammaPrior(2.0, 0.44),
                "eta": 0.6,
            }
        }
        model = BotorchModel(**kwargs)
        datasets = [
            SupervisedDataset(
                X=Xs1[0],
                Y=Ys1[0],
                Yvar=Yvars1[0],
                feature_names=feature_names,
                outcome_names=metric_names,
            ),
            SupervisedDataset(
                X=Xs2[0],
                Y=Ys2[0],
                Yvar=Yvars2[0],
                feature_names=feature_names,
                outcome_names=metric_names,
            ),
        ]

        with mock.patch(FIT_MODEL_MO_PATH) as _mock_fit_model:
            model.fit(
                datasets=datasets,
                search_space_digest=SearchSpaceDigest(
                    feature_names=feature_names,
                    bounds=bounds,
                    task_features=[0],
                ),
            )
            _mock_fit_model.assert_called_once()

        # Check ranks
        model_list = cast(ModelListGP, model.model).models
        for i in range(1):
            self.assertEqual(
                model_list[i].covar_module.base_kernel.lengthscale_prior.concentration,
                6.0,
            )
            self.assertEqual(
                model_list[i].covar_module.base_kernel.lengthscale_prior.rate,
                3.0,
            )
            self.assertEqual(
                model_list[i].covar_module.outputscale_prior.concentration,
                3.0,
            )
            self.assertEqual(
                model_list[i].covar_module.outputscale_prior.rate,
                12.0,
            )
            self.assertIsInstance(
                model_list[i].task_covar_module.IndexKernelPrior, LKJCovariancePrior
            )
            self.assertEqual(
                model_list[i].task_covar_module.IndexKernelPrior.sd_prior.concentration,
                2.0,
            )
            self.assertEqual(
                model_list[i].task_covar_module.IndexKernelPrior.sd_prior.rate, 0.44
            )
            self.assertEqual(
                model_list[i].task_covar_module.IndexKernelPrior.correlation_prior.eta,
                0.6,
            )

    @fast_botorch_optimize
    def test_BotorchModel(
        self, dtype: torch.dtype = torch.float, cuda: bool = False
    ) -> None:
        (
            Xs1,
            Ys1,
            Yvars1,
            bounds,
            tfs,
            feature_names,
            metric_names,
        ) = get_torch_test_data(dtype=dtype, cuda=cuda, constant_noise=True)
        Xs2, Ys2, Yvars2, _, _, _, _ = get_torch_test_data(
            dtype=dtype, cuda=cuda, constant_noise=True
        )
        for use_input_warping in (True, False):
            for use_loocv_pseudo_likelihood in (True, False):
                model = BotorchModel(
                    use_input_warping=use_input_warping,
                    use_loocv_pseudo_likelihood=use_loocv_pseudo_likelihood,
                )

                # Test ModelListGP

                # make training data different for each output
                Xs2_diff = [Xs2[0] + 0.1]
                datasets = [
                    SupervisedDataset(
                        X=Xs1[0],
                        Y=Ys1[0],
                        Yvar=Yvars1[0],
                        feature_names=feature_names,
                        outcome_names=metric_names,
                    ),
                    SupervisedDataset(
                        X=Xs2_diff[0],
                        Y=Ys2[0],
                        Yvar=Yvars2[0],
                        feature_names=feature_names,
                        outcome_names=metric_names,
                    ),
                ]
                with mock.patch(FIT_MODEL_MO_PATH) as _mock_fit_model:
                    model.fit(
                        datasets=datasets,
                        search_space_digest=SearchSpaceDigest(
                            feature_names=feature_names,
                            bounds=bounds,
                            task_features=tfs,
                        ),
                    )
                    _mock_fit_model.assert_called_once()
                    if use_loocv_pseudo_likelihood:
                        mll_cls = LeaveOneOutPseudoLikelihood
                    else:
                        mll_cls = ExactMarginalLogLikelihood
                    mlls = _mock_fit_model.mock_calls[0][1][0].mlls
                    self.assertTrue(len(mlls) == 2)
                    for mll in mlls:
                        self.assertIsInstance(mll, mll_cls)
                # Check attributes
                self.assertTrue(torch.equal(model.Xs[0], Xs1[0]))
                self.assertTrue(torch.equal(model.Xs[1], Xs2_diff[0]))
                self.assertEqual(model.dtype, Xs1[0].dtype)
                self.assertEqual(model.device, Xs1[0].device)
                self.assertIsInstance(model.model, ModelListGP)

                # Check fitting
                model_list = cast(ModelListGP, model.model).models
                self.assertTrue(torch.equal(model_list[0].train_inputs[0], Xs1[0]))
                self.assertTrue(torch.equal(model_list[1].train_inputs[0], Xs2_diff[0]))
                self.assertTrue(
                    torch.equal(model_list[0].train_targets, Ys1[0].view(-1))
                )
                self.assertTrue(
                    torch.equal(model_list[1].train_targets, Ys2[0].view(-1))
                )
                self.assertIsInstance(model_list[0].likelihood, _GaussianLikelihoodBase)
                self.assertIsInstance(model_list[1].likelihood, _GaussianLikelihoodBase)
                if use_input_warping:
                    self.assertTrue(model.use_input_warping)
                for m in model_list:
                    if use_input_warping:
                        self.assertTrue(hasattr(m, "input_transform"))
                        self.assertIsInstance(m.input_transform, Warp)
                    else:
                        self.assertFalse(hasattr(m, "input_transform"))

            # Test batched multi-output SingleTaskGP
            datasets_block = [
                SupervisedDataset(
                    X=Xs1[0],
                    Y=Ys1[0],
                    Yvar=Yvars1[0],
                    feature_names=feature_names,
                    outcome_names=metric_names,
                ),
                SupervisedDataset(
                    X=Xs2[0],
                    Y=Ys2[0],
                    Yvar=Yvars2[0],
                    feature_names=feature_names,
                    outcome_names=metric_names,
                ),
            ]
            with mock.patch(FIT_MODEL_MO_PATH) as _mock_fit_model:
                model.fit(
                    datasets=datasets_block,
                    search_space_digest=SearchSpaceDigest(
                        feature_names=feature_names,
                        bounds=bounds,
                        task_features=tfs,
                    ),
                )
                _mock_fit_model.assert_called_once()

            # Check attributes
            self.assertTrue(torch.equal(model.Xs[0], Xs1[0]))
            self.assertTrue(torch.equal(model.Xs[1], Xs2[0]))
            self.assertEqual(model.dtype, Xs1[0].dtype)
            self.assertEqual(model.device, Xs1[0].device)
            if use_input_warping:
                self.assertIsInstance(model.model, ModelListGP)
                models = model.model.models
            else:
                models = [model.model]
            Ys = [Ys1[0], Ys2[0]]
            for i, m in enumerate(models):
                self.assertIsInstance(m, SingleTaskGP)
                self.assertIsInstance(m.likelihood, FixedNoiseGaussianLikelihood)
                expected_train_inputs = Xs1[0]

                if not use_input_warping:
                    expected_train_inputs = expected_train_inputs.unsqueeze(0).expand(
                        2, *Xs1[0].shape
                    )
                    expected_train_targets = torch.cat(Ys1 + Ys2, dim=-1).permute(1, 0)
                else:
                    expected_train_targets = Ys[i].squeeze(-1)
                # Check fitting
                # train inputs should be `o x n x 1`
                self.assertTrue(
                    torch.equal(
                        m.train_inputs[0],
                        expected_train_inputs,
                    )
                )
                # train targets should be `o x n`
                self.assertTrue(
                    torch.equal(
                        m.train_targets,
                        expected_train_targets,
                    )
                )
                self.assertIsInstance(m.likelihood, _GaussianLikelihoodBase)

            # Check infeasible cost can be computed on the model
            tkwargs: Dict[str, Any] = {
                "device": torch.device("cuda" if cuda else "cpu"),
                "dtype": dtype,
            }
            objective_weights = torch.tensor([1.0, 0.0], **tkwargs)
            objective_transform = get_objective_weights_transform(objective_weights)
            infeasible_cost = torch.tensor(
                get_infeasible_cost(
                    X=Xs1[0], model=model.model, objective=objective_transform
                )
            )
            expected_infeasible_cost = -1 * torch.min(
                # pyre-fixme[20]: Argument `1` expected.
                objective_transform(
                    model.model.posterior(Xs1[0]).mean
                    - 6 * model.model.posterior(Xs1[0]).variance.sqrt()
                ).min(),
                torch.tensor(0.0, **tkwargs),
            )
            self.assertTrue(
                torch.abs(infeasible_cost - expected_infeasible_cost) < 1e-5
            )

            # Check prediction
            X = torch.tensor([[6.0, 7.0, 8.0]], **tkwargs)
            f_mean, f_cov = model.predict(X)
            self.assertTrue(f_mean.shape == torch.Size([1, 2]))
            self.assertTrue(f_cov.shape == torch.Size([1, 2, 2]))

            # Check generation
            objective_weights = torch.tensor([1.0, 0.0], **tkwargs)
            outcome_constraints = (
                torch.tensor([[0.0, 1.0]], **tkwargs),
                torch.tensor([[5.0]], **tkwargs),
            )
            linear_constraints = (
                torch.tensor([[0.0, 1.0, 1.0]]),
                torch.tensor([[100.0]]),
            )
            fixed_features = None
            pending_observations = [
                torch.tensor([[1.0, 3.0, 4.0]], **tkwargs),
                torch.tensor([[2.0, 6.0, 8.0]], **tkwargs),
            ]
            n = 3

            X_dummy = torch.tensor([[[1.0, 2.0, 3.0]]], **tkwargs)
            acqfv_dummy = torch.tensor([[[1.0, 2.0, 3.0]]], **tkwargs)
            model_gen_options = {"subset_model": False}
            # test sequential optimize
            search_space_digest = SearchSpaceDigest(
                feature_names=[],
                bounds=bounds,
            )
            with mock.patch(
                "ax.models.torch.botorch_defaults.optimize_acqf",
                return_value=(X_dummy, acqfv_dummy),
            ) as mock_optimize_acqf:
                gen_results = model.gen(
                    n=n,
                    search_space_digest=search_space_digest,
                    torch_opt_config=TorchOptConfig(
                        objective_weights=objective_weights,
                        outcome_constraints=outcome_constraints,
                        linear_constraints=linear_constraints,
                        fixed_features=fixed_features,
                        pending_observations=pending_observations,
                        # pyre-fixme[6]: For 6th param expected `Dict[str,
                        #  Union[None, Dict[str, typing.Any], OptimizationConfig,
                        #  AcquisitionFunction, float, int, str]]` but got `Dict[str,
                        #  bool]`.
                        model_gen_options=model_gen_options,
                        rounding_func=dummy_func,
                    ),
                )
                # note: gen() always returns CPU tensors
                self.assertTrue(torch.equal(gen_results.points, X_dummy.cpu()))
                self.assertTrue(
                    torch.equal(gen_results.weights, torch.ones(n, dtype=dtype))
                )
            self.assertEqual(
                mock_optimize_acqf.call_args.kwargs["options"]["init_batch_limit"], 32
            )
            self.assertEqual(
                mock_optimize_acqf.call_args.kwargs["options"]["batch_limit"], 5
            )
            torch_opt_config = TorchOptConfig(
                objective_weights=objective_weights,
                fixed_features=fixed_features,
                pending_observations=pending_observations,
                model_gen_options={"optimizer_kwargs": {"joint_optimization": True}},
            )
            # test joint optimize
            with mock.patch(
                "ax.models.torch.botorch_defaults.optimize_acqf",
                return_value=(X_dummy, acqfv_dummy),
            ) as mock_optimize_acqf:
                gen_results = model.gen(
                    n=n,
                    search_space_digest=search_space_digest,
                    torch_opt_config=torch_opt_config,
                )
                # note: gen() always returns CPU tensors
                self.assertTrue(torch.equal(gen_results.points, X_dummy.cpu()))
                self.assertTrue(
                    torch.equal(gen_results.weights, torch.ones(n, dtype=dtype))
                )
                mock_optimize_acqf.assert_called_once()

            # test that fidelity features are unsupported
            with self.assertRaises(NotImplementedError):
                gen_results = model.gen(
                    n=n,
                    search_space_digest=dataclasses.replace(
                        search_space_digest,
                        fidelity_features=[0],
                        target_values={0: 3.0},
                    ),
                    torch_opt_config=torch_opt_config,
                )

            # test get_rounding_func
            dummy_rounding = not_none(get_rounding_func(rounding_func=dummy_func))
            X_temp = torch.rand(1, 2, 3, 4)
            self.assertTrue(torch.equal(X_temp, dummy_rounding(X_temp)))

            # Check best point selection
            xbest = model.best_point(
                search_space_digest=search_space_digest,
                torch_opt_config=torch_opt_config,
            )
            torch_opt_config = dataclasses.replace(
                torch_opt_config,
                fixed_features={0: 100.0},
            )
            xbest = model.best_point(
                search_space_digest=search_space_digest,
                torch_opt_config=torch_opt_config,
            )
            self.assertIsNone(xbest)

            # test that fidelity features are unsupported
            with self.assertRaises(NotImplementedError):
                xbest = model.best_point(
                    search_space_digest=dataclasses.replace(
                        search_space_digest,
                        fidelity_features=[0],
                        target_values={0: 3.0},
                    ),
                    torch_opt_config=torch_opt_config,
                )

            # Test cross-validation
            combined_datasets = [
                SupervisedDataset(
                    Xs1[0],
                    Y=Ys1[0],
                    Yvar=Yvars1[0],
                    feature_names=feature_names,
                    outcome_names=metric_names,
                ),
                SupervisedDataset(
                    Xs2[0],
                    Y=Ys2[0],
                    Yvar=Yvars2[0],
                    feature_names=feature_names,
                    outcome_names=metric_names,
                ),
            ]
            mean, variance = model.cross_validate(
                datasets=combined_datasets,
                X_test=torch.tensor([[1.2, 3.2, 4.2], [2.4, 5.2, 3.2]], **tkwargs),
            )
            self.assertTrue(mean.shape == torch.Size([2, 2]))
            self.assertTrue(variance.shape == torch.Size([2, 2, 2]))

            # Test cross-validation with refit_on_cv
            model.refit_on_cv = True
            mean, variance = model.cross_validate(
                datasets=combined_datasets,
                X_test=torch.tensor([[1.2, 3.2, 4.2], [2.4, 5.2, 3.2]], **tkwargs),
            )
            self.assertTrue(mean.shape == torch.Size([2, 2]))
            self.assertTrue(variance.shape == torch.Size([2, 2, 2]))

            # Test feature_importances
            importances = model.feature_importances()
            self.assertEqual(importances.shape, torch.Size([2, 1, 3]))

            # test unfit model CV and feature_importances
            unfit_model = BotorchModel()
            with self.assertRaisesRegex(
                RuntimeError, r"Cannot cross-validate model that has not been fitted"
            ):
                unfit_model.cross_validate(
                    datasets=combined_datasets,
                    X_test=Xs1[0],
                )
            with self.assertRaisesRegex(
                RuntimeError,
                r"Cannot calculate feature_importances without a fitted model",
            ):
                unfit_model.feature_importances()

            # Test loading state dict
            true_state_dict = {
                "mean_module.raw_constant": 3.5004,
                "covar_module.raw_outputscale": 2.2438,
                "covar_module.base_kernel.raw_lengthscale": [
                    [-0.9274, -0.9274, -0.9274]
                ],
                "covar_module.base_kernel.raw_lengthscale_constraint.lower_bound": 0.1,
                "covar_module.base_kernel.raw_lengthscale_constraint.upper_bound": 2.5,
                "covar_module.base_kernel.lengthscale_prior.concentration": 3.0,
                "covar_module.base_kernel.lengthscale_prior.rate": 6.0,
                "covar_module.raw_outputscale_constraint.lower_bound": 0.2,
                "covar_module.raw_outputscale_constraint.upper_bound": 2.6,
                "covar_module.outputscale_prior.concentration": 2.0,
                "covar_module.outputscale_prior.rate": 0.15,
            }
            true_state_dict = {
                key: torch.tensor(val, **tkwargs)
                for key, val in true_state_dict.items()
            }
            model = get_and_fit_model(
                Xs=Xs1,
                Ys=Ys1,
                Yvars=Yvars1,
                task_features=[],
                fidelity_features=[],
                metric_names=[metric_names[0]],
                state_dict=true_state_dict,
                refit_model=False,
            )
            for k, v in chain(model.named_parameters(), model.named_buffers()):
                self.assertTrue(torch.equal(true_state_dict[k], v))

            # Test for some change in model parameters & buffer for refit_model=True
            true_state_dict["mean_module.raw_constant"] += 0.1
            true_state_dict["covar_module.raw_outputscale"] += 0.1
            true_state_dict["covar_module.base_kernel.raw_lengthscale"] += 0.1
            model = get_and_fit_model(
                Xs=Xs1,
                Ys=Ys1,
                Yvars=Yvars1,
                task_features=[],
                fidelity_features=[],
                metric_names=[metric_names[0]],
                state_dict=true_state_dict,
                refit_model=True,
            )
            self.assertTrue(
                any(
                    not torch.equal(true_state_dict[k], v)
                    for k, v in chain(model.named_parameters(), model.named_buffers())
                )
            )

        # Test that recommend_best_out_of_sample_point errors w/o _get_best_point_acqf
        model = BotorchModel(best_point_recommender=recommend_best_out_of_sample_point)
        with mock.patch(FIT_MODEL_MO_PATH) as _mock_fit_model:
            model.fit(
                # pyre-fixme[61]: `datasets` is undefined, or not always defined.
                datasets=datasets,
                search_space_digest=SearchSpaceDigest(
                    feature_names=feature_names,
                    bounds=bounds,
                    task_features=tfs,
                ),
            )
        with self.assertRaises(RuntimeError):
            xbest = model.best_point(
                # pyre-fixme[61]: `search_space_digest` is undefined, or not always
                #  defined.
                search_space_digest=search_space_digest,
                # pyre-fixme[61]: `torch_opt_config` is undefined, or not always
                #  defined.
                torch_opt_config=torch_opt_config,
            )

    def test_BotorchModel_cuda(self) -> None:
        if torch.cuda.is_available():
            self.test_BotorchModel(cuda=True)

    def test_BotorchModel_double(self) -> None:
        self.test_BotorchModel(dtype=torch.double)

    def test_BotorchModel_double_cuda(self) -> None:
        if torch.cuda.is_available():
            self.test_BotorchModel(dtype=torch.double, cuda=True)

    def test_BotorchModelOneOutcome(self) -> None:
        (
            Xs1,
            Ys1,
            Yvars1,
            bounds,
            tfs,
            feature_names,
            metric_names,
        ) = get_torch_test_data(dtype=torch.float, cuda=False, constant_noise=True)
        for use_input_warping, use_loocv_pseudo_likelihood in product(
            (True, False), (True, False)
        ):
            model = BotorchModel(
                use_input_warping=use_input_warping,
                use_loocv_pseudo_likelihood=use_loocv_pseudo_likelihood,
            )
            with mock.patch(FIT_MODEL_MO_PATH) as _mock_fit_model:
                model.fit(
                    datasets=[
                        SupervisedDataset(
                            X=Xs1[0],
                            Y=Ys1[0],
                            Yvar=Yvars1[0],
                            feature_names=feature_names,
                            outcome_names=metric_names,
                        )
                    ],
                    search_space_digest=SearchSpaceDigest(
                        feature_names=feature_names,
                        bounds=bounds,
                        task_features=tfs,
                    ),
                )
                _mock_fit_model.assert_called_once()
                if use_loocv_pseudo_likelihood:
                    mll_cls = LeaveOneOutPseudoLikelihood
                else:
                    mll_cls = ExactMarginalLogLikelihood
                self.assertIsInstance(
                    _mock_fit_model.mock_calls[0][1][0],
                    mll_cls,
                )
            X = torch.rand(2, 3, dtype=torch.float)
            f_mean, f_cov = model.predict(X)
            self.assertTrue(f_mean.shape == torch.Size([2, 1]))
            self.assertTrue(f_cov.shape == torch.Size([2, 1, 1]))
            if use_input_warping:
                self.assertTrue(hasattr(model.model, "input_transform"))
                self.assertIsInstance(model.model.input_transform, Warp)
            else:
                self.assertFalse(hasattr(model.model, "input_transform"))

    def test_BotorchModelConstraints(self) -> None:
        (
            Xs1,
            Ys1,
            Yvars1,
            bounds,
            tfs,
            feature_names,
            metric_names,
        ) = get_torch_test_data(dtype=torch.float, cuda=False, constant_noise=True)
        Xs2, Ys2, Yvars2, _, _, _, _ = get_torch_test_data(
            dtype=torch.float, cuda=False, constant_noise=True
        )
        # make infeasible
        Xs2[0] = -1 * Xs2[0]
        objective_weights = torch.tensor(
            [-1.0, 1.0], dtype=torch.float, device=torch.device("cpu")
        )
        n = 3
        model = BotorchModel()
        search_space_digest = SearchSpaceDigest(
            feature_names=feature_names,
            bounds=bounds,
            task_features=tfs,
        )
        with mock.patch(FIT_MODEL_MO_PATH) as _mock_fit_model:
            model.fit(
                datasets=[
                    SupervisedDataset(
                        X=Xs1[0],
                        Y=Ys1[0],
                        Yvar=Yvars1[0],
                        feature_names=feature_names,
                        outcome_names=metric_names,
                    ),
                    SupervisedDataset(
                        X=Xs2[0],
                        Y=Ys2[0],
                        Yvar=Yvars2[0],
                        feature_names=feature_names,
                        outcome_names=metric_names,
                    ),
                ],
                search_space_digest=search_space_digest,
            )
            _mock_fit_model.assert_called_once()

        # because there are no feasible points:
        with self.assertRaises(ValueError):
            model.gen(
                n,
                search_space_digest=search_space_digest,
                torch_opt_config=TorchOptConfig(objective_weights),
            )

    def test_botorchmodel_raises_when_no_data(self) -> None:
        _, _, _, bounds, tfs, feature_names, metric_names = get_torch_test_data(
            dtype=torch.float, cuda=False, constant_noise=True
        )
        search_space_digest = SearchSpaceDigest(
            feature_names=feature_names,
            bounds=bounds,
            task_features=tfs,
        )
        model = BotorchModel()
        with self.assertRaisesRegex(
            DataRequiredError, "BotorchModel.fit requires non-empty data sets."
        ):
            model.fit(
                datasets=[],
                search_space_digest=search_space_digest,
            )

    def test_get_feature_importances_from_botorch_model(self) -> None:
        tkwargs: Dict[str, Any] = {"dtype": torch.double}
        train_X = torch.rand(5, 3, **tkwargs)
        train_Y = train_X.sum(dim=-1, keepdim=True)
        simple_gp = SingleTaskGP(train_X=train_X, train_Y=train_Y)
        simple_gp.covar_module.base_kernel.lengthscale = torch.tensor(
            [1, 3, 5], **tkwargs
        )
        importances = get_feature_importances_from_botorch_model(simple_gp)
        self.assertTrue(np.allclose(importances, np.array([15 / 23, 5 / 23, 3 / 23])))
        self.assertEqual(importances.shape, (1, 1, 3))
        # Model with no base kernel
        simple_gp.covar_module.base_kernel = None
        with self.assertRaisesRegex(
            NotImplementedError,
            "Failed to extract lengthscales from `m.covar_module.base_kernel`",
        ):
            get_feature_importances_from_botorch_model(simple_gp)

        with self.assertRaisesRegex(
            RuntimeError,
            "Cannot calculate feature_importances without a fitted model",
        ):
            get_feature_importances_from_botorch_model(None)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from __future__ import annotations

from dataclasses import replace as dataclass_replace

from logging import Logger
from typing import Dict, Optional, Set

from ax.core.experiment import Experiment
from ax.modelbridge.generation_strategy import GenerationStrategy
from ax.runners.simulated_backend import SimulatedBackendRunner
from ax.service.scheduler import Scheduler, SchedulerOptions
from ax.utils.common.logger import get_logger
from ax.utils.testing.backend_simulator import BackendSimulator

logger: Logger = get_logger(__name__)


class AsyncSimulatedBackendScheduler(Scheduler):
    """A Scheduler that uses a simulated backend for Ax asynchronous benchmarks."""

    def __init__(
        self,
        experiment: Experiment,
        generation_strategy: GenerationStrategy,
        max_pending_trials: int,
        options: SchedulerOptions,
    ) -> None:
        """A Scheduler for Ax asynchronous benchmarks.

        Args:
            experiment: Experiment, in which results of the optimization
                will be recorded.
            generation_strategy: Generation strategy for the optimization,
                describes models that will be used in optimization.
            max_pending_trials: The maximum number of pending trials allowed.
            options: `SchedulerOptions` for this Scheduler instance.
        """
        if not isinstance(experiment.runner, SimulatedBackendRunner):
            raise ValueError(
                "experiment must have runner of type SimulatedBackendRunner attached"
            )
        if (
            options.max_pending_trials is not None
            and options.max_pending_trials != max_pending_trials
        ):
            raise ValueError(
                f"`SchedulerOptions.max_pending_trials`: {options.max_pending_trials} "
                f"does not match argument to `Scheduler`: {max_pending_trials}."
            )
        if options.max_pending_trials is None:
            options = dataclass_replace(options, max_pending_trials=max_pending_trials)

        super().__init__(
            experiment=experiment,
            generation_strategy=generation_strategy,
            options=options,
            _skip_experiment_save=True,
        )

    @property
    def backend_simulator(self) -> BackendSimulator:
        """Get the ``BackendSimulator`` stored on the runner of the experiment.

        Returns:
            The backend simulator.
        """
        return self.runner.simulator  # pyre-ignore[16]

    def should_stop_trials_early(
        self, trial_indices: Set[int]
    ) -> Dict[int, Optional[str]]:
        """Given a set of trial indices, decide whether or not to early-stop
        running trials using the ``early_stopping_strategy``.

        Args:
            trial_indices: Indices of trials to consider for early stopping.

        Returns:
            Dict with new suggested ``TrialStatus`` as keys and a set of
            indices of trials to update (subset of initially-passed trials) as values.
        """
        # TODO: The status on the experiment does not distinguish between
        # running and queued trials, so here we check status on the
        # ``backend_simulator`` directly to make sure it is running.
        running_trials = set()
        skipped_trials = set()
        for trial_index in trial_indices:
            sim_trial = self.backend_simulator.get_sim_trial_by_index(trial_index)
            if sim_trial.sim_start_time is not None and (  # pyre-ignore[16]
                self.backend_simulator.time - sim_trial.sim_start_time > 0
            ):
                running_trials.add(trial_index)
            else:
                skipped_trials.add(trial_index)
        if len(skipped_trials) > 0:
            logger.info(
                f"Not sending {skipped_trials} to base `should_stop_trials_early` "
                "because they have not been running for a positive amount of time "
                "on the backend simulator."
            )
        return super().should_stop_trials_early(trial_indices=running_trials)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from __future__ import annotations

from typing import List, Optional, Tuple

import numpy as np
import torch
from ax.core.search_space import SearchSpaceDigest
from ax.core.types import TCandidateMetadata
from ax.models.torch.utils import _datasets_to_legacy_inputs
from ax.models.torch_base import TorchModel
from ax.utils.common.docutils import copy_doc
from botorch.utils.datasets import SupervisedDataset
from sklearn.ensemble import RandomForestRegressor
from sklearn.tree import DecisionTreeRegressor
from torch import Tensor


class RandomForest(TorchModel):
    """A Random Forest model.

    Uses a parametric bootstrap to handle uncertainty in Y.

    Can be used to fit data, make predictions, and do cross validation; however
    gen is not implemented and so this model cannot generate new points.

    Args:
        max_features: Maximum number of features at each split. With one-hot
            encoding, this should be set to None. Defaults to "sqrt", which is
            Breiman's version of Random Forest.
        num_trees: Number of trees.
    """

    def __init__(
        self, max_features: Optional[str] = "sqrt", num_trees: int = 500
    ) -> None:
        self.max_features = max_features
        self.num_trees = num_trees
        self.models: List[RandomForestRegressor] = []

    @copy_doc(TorchModel.fit)
    def fit(
        self,
        datasets: List[SupervisedDataset],
        search_space_digest: SearchSpaceDigest,
        candidate_metadata: Optional[List[List[TCandidateMetadata]]] = None,
    ) -> None:
        Xs, Ys, Yvars = _datasets_to_legacy_inputs(datasets=datasets)
        for X, Y, Yvar in zip(Xs, Ys, Yvars):
            self.models.append(
                _get_rf(
                    X=X.numpy(),
                    Y=Y.numpy(),
                    Yvar=Yvar.numpy(),
                    num_trees=self.num_trees,
                    max_features=self.max_features,
                )
            )

    @copy_doc(TorchModel.predict)
    def predict(self, X: Tensor) -> Tuple[Tensor, Tensor]:
        return _rf_predict(self.models, X)

    @copy_doc(TorchModel.cross_validate)
    def cross_validate(  # pyre-ignore [14]: not using metric_names or ssd
        self,
        datasets: List[SupervisedDataset],
        X_test: Tensor,
        use_posterior_predictive: bool = False,
    ) -> Tuple[Tensor, Tensor]:
        Xs, Ys, Yvars = _datasets_to_legacy_inputs(datasets=datasets)
        cv_models: List[RandomForestRegressor] = []
        for X, Y, Yvar in zip(Xs, Ys, Yvars):
            cv_models.append(
                _get_rf(
                    X=X.numpy(),
                    Y=Y.numpy(),
                    Yvar=Yvar.numpy(),
                    num_trees=self.num_trees,
                    max_features=self.max_features,
                )
            )
        return _rf_predict(cv_models, X_test)


def _get_rf(
    X: np.ndarray,
    Y: np.ndarray,
    Yvar: np.ndarray,
    num_trees: int,
    max_features: Optional[str],
) -> RandomForestRegressor:
    """Fit a Random Forest model.

    Args:
        X: X
        Y: Y
        Yvar: Variance for Y
        num_trees: Number of trees
        max_features: Max features specifier

    Returns: Fitted Random Forest.
    """
    r = RandomForestRegressor(
        n_estimators=num_trees, max_features=max_features, bootstrap=True
    )
    # pyre-fixme[16]: `RandomForestRegressor` has no attribute `estimators_`.
    r.estimators_ = [DecisionTreeRegressor() for i in range(r.n_estimators)]
    for estimator in r.estimators_:
        # Parametric bootstrap
        y = np.random.normal(loc=Y[:, 0], scale=np.sqrt(Yvar[:, 0]))
        estimator.fit(X, y)
    return r


def _rf_predict(
    models: List[RandomForestRegressor], X: Tensor
) -> Tuple[Tensor, Tensor]:
    """Make predictions with Random Forest models.

    Args:
        models: List of models for each outcome
        X: X to predict

    Returns:
        mean and covariance estimates
    """
    f = np.zeros((X.shape[0], len(models)))
    cov = np.zeros((X.shape[0], len(models), len(models)))
    for i, m in enumerate(models):
        # pyre-fixme[16]: `RandomForestRegressor` has no attribute `estimators_`.
        preds = np.vstack([tree.predict(X.numpy()) for tree in m.estimators_])
        f[:, i] = preds.mean(0)
        cov[:, i, i] = preds.var(0)
    return torch.from_numpy(f), torch.from_numpy(cov)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict


import numpy as np
from ax.models.discrete.full_factorial import FullFactorialGenerator
from ax.utils.common.testutils import TestCase


class FullFactorialGeneratorTest(TestCase):
    def test_FullFactorial(self) -> None:
        generator = FullFactorialGenerator()
        parameter_values = [[1, 2], ["foo", "bar"]]
        generated_points, weights, _ = generator.gen(
            n=-1,
            # pyre-fixme[6]: For 2nd param expected `List[List[Union[None, bool,
            #  float, int, str]]]` but got `List[Union[List[int], List[str]]]`.
            parameter_values=parameter_values,
            objective_weights=np.ones(1),
        )
        expected_points = [[1, "foo"], [1, "bar"], [2, "foo"], [2, "bar"]]
        self.assertEqual(generated_points, expected_points)
        self.assertEqual(weights, [1 for _ in range(len(expected_points))])

    def test_FullFactorialValidation(self) -> None:
        # Raise error because cardinality exceeds max cardinality
        generator = FullFactorialGenerator(max_cardinality=5, check_cardinality=True)
        parameter_values = [[1, 2], ["foo", "bar"], [True, False]]
        with self.assertRaises(ValueError):
            generated_points, weights, _ = generator.gen(
                n=-1,
                # pyre-fixme[6]: For 2nd param expected `List[List[Union[None, bool,
                #  float, int, str]]]` but got `List[Union[List[bool], List[int],
                #  List[str]]]`.
                parameter_values=parameter_values,
                objective_weights=np.ones(1),
            )

        # Raise error because n != -1
        generator = FullFactorialGenerator()
        parameter_values = [[1, 2], ["foo", "bar"]]
        with self.assertRaises(ValueError):
            generated_points, weights, _ = generator.gen(
                n=5,
                # pyre-fixme[6]: For 2nd param expected `List[List[Union[None, bool,
                #  float, int, str]]]` but got `List[Union[List[int], List[str]]]`.
                parameter_values=parameter_values,
                objective_weights=np.ones(1),
            )

    def test_FullFactorialFixedFeatures(self) -> None:
        generator = FullFactorialGenerator(max_cardinality=5, check_cardinality=True)
        parameter_values = [[1, 2], ["foo", "bar"]]
        generated_points, weights, _ = generator.gen(
            n=-1,
            # pyre-fixme[6]: For 2nd param expected `List[List[Union[None, bool,
            #  float, int, str]]]` but got `List[Union[List[int], List[str]]]`.
            parameter_values=parameter_values,
            objective_weights=np.ones(1),
            fixed_features={1: "foo"},
        )
        expected_points = [[1, "foo"], [2, "foo"]]
        self.assertEqual(generated_points, expected_points)
        self.assertEqual(weights, [1 for _ in range(len(expected_points))])



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from __future__ import annotations

from abc import ABC, abstractmethod
from enum import Enum
from functools import partial
from typing import Any, Callable, Dict, List, Optional, Union

import numpy as np
from ax.exceptions.core import UserInputError
from ax.modelbridge.model_spec import ModelSpec
from ax.utils.common.base import Base
from ax.utils.common.typeutils import not_none

ARRAYLIKE = Union[np.ndarray, List[float], List[np.ndarray]]


class BestModelSelector(ABC, Base):
    @abstractmethod
    def best_model(self, model_specs: List[ModelSpec]) -> int:
        """
        Return the index of the best ``ModelSpec``.
        """


class ReductionCriterion(Enum):
    """An enum for callables that are used for aggregating diagnostics over metrics
    and selecting the best diagnostic in ``SingleDiagnosticBestModelSelector``.

    NOTE: This is used to ensure serializability of the callables.
    """

    # NOTE: Callables need to be wrapped in `partial` to be registered as members.
    MEAN: Callable[[ARRAYLIKE], np.ndarray] = partial(np.mean)
    MIN: Callable[[ARRAYLIKE], np.ndarray] = partial(np.min)
    MAX: Callable[[ARRAYLIKE], np.ndarray] = partial(np.max)

    def __call__(self, array_like: ARRAYLIKE) -> np.ndarray:
        return self.value(array_like)


class SingleDiagnosticBestModelSelector(BestModelSelector):
    """Choose the best model using a single cross-validation diagnostic.

    The input is a list of ``ModelSpec``, each corresponding to one model.
    The specified diagnostic is extracted from each of the models,
    its values (each of which corresponds to a separate metric) are
    aggregated with the aggregation function, the best one is determined
    with the criterion, and the index of the best diagnostic result is returned.

    Example:
     ::
        s = SingleDiagnosticBestModelSelector(
            diagnostic='Fisher exact test p',
            metric_aggregation=ReductionCriterion.MEAN,
            criterion=ReductionCriterion.MIN,
            model_cv_kwargs={"untransform": False},
        )
        best_diagnostic_index = s.best_diagnostic(diagnostics)

    Args:
        diagnostic: The name of the diagnostic to use, which should be
            a key in ``CVDiagnostic``.
        metric_aggregation: ``ReductionCriterion`` applied to the values of the
            diagnostic for a single model to produce a single number.
        criterion: ``ReductionCriterion`` used to determine which of the
            (aggregated) diagnostics is the best.
        model_cv_kwargs: Optional dictionary of kwargs to pass in while computing
            the cross validation diagnostics.

    Returns:
        int: index of the selected best diagnostic.
    """

    def __init__(
        self,
        diagnostic: str,
        metric_aggregation: ReductionCriterion,
        criterion: ReductionCriterion,
        model_cv_kwargs: Optional[Dict[str, Any]] = None,
    ) -> None:
        self.diagnostic = diagnostic
        if not isinstance(metric_aggregation, ReductionCriterion) or not isinstance(
            criterion, ReductionCriterion
        ):
            raise UserInputError(
                "Both `metric_aggregation` and `criterion` must be "
                f"`ReductionCriterion`. Got {metric_aggregation=}, {criterion=}."
            )
        if criterion == ReductionCriterion.MEAN:
            raise UserInputError(
                f"{criterion=} is not supported. Please use MIN or MAX."
            )
        self.metric_aggregation = metric_aggregation
        self.criterion = criterion
        self.model_cv_kwargs = model_cv_kwargs

    def best_model(self, model_specs: List[ModelSpec]) -> int:
        for model_spec in model_specs:
            model_spec.cross_validate(model_cv_kwargs=self.model_cv_kwargs)
        aggregated_diagnostic_values = [
            self.metric_aggregation(
                list(not_none(model_spec.diagnostics)[self.diagnostic].values())
            )
            for model_spec in model_specs
        ]
        best_diagnostic = self.criterion(aggregated_diagnostic_values).item()
        return aggregated_diagnostic_values.index(best_diagnostic)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import math
import random
from copy import copy
from typing import Set

import numpy as np
from ax.core.parameter_constraint import OrderConstraint
from ax.core.search_space import SearchSpace
from ax.core.types import TParameterization


def randomized_round(x: float) -> int:
    """Randomized round of x"""
    z = math.floor(x)
    return int(z + float(random.random() <= (x - z)))


def randomized_onehot_round(x: np.ndarray) -> np.ndarray:
    """Randomized rounding of x to a one-hot vector.
    x should be 0 <= x <= 1. If x includes negative values,
    they will be rounded to zero.
    """
    neg_x = x < 0
    x[neg_x] = 0
    if len(x) == 1:
        return np.array([randomized_round(x[0])])
    if sum(x) == 0:
        x = np.ones_like(x)
        x[neg_x] = 0
    w = x / sum(x)
    hot = np.random.choice(len(w), size=1, p=w)[0]
    z = np.zeros_like(x)
    z[hot] = 1
    return z


def strict_onehot_round(x: np.ndarray) -> np.ndarray:
    """Round x to a one-hot vector by selecting the max element.
    Ties broken randomly."""
    if len(x) == 1:
        return np.round(x)
    argmax = x == max(x)
    x[argmax] = 1
    x[~argmax] = 0
    return randomized_onehot_round(x)


def contains_constrained_integer(
    search_space: SearchSpace, transform_parameters: Set[str]
) -> bool:
    """Check if any integer parameters are present in parameter_constraints.

    Order constraints are ignored since strict rounding preserves ordering.
    """
    for constraint in search_space.parameter_constraints:
        if isinstance(constraint, OrderConstraint):
            continue
        constraint_params = set(constraint.constraint_dict.keys())
        if constraint_params.intersection(transform_parameters):
            return True
    return False


def randomized_round_parameters(
    parameters: TParameterization, transform_parameters: Set[str]
) -> TParameterization:
    rounded_parameters = copy(parameters)
    for p_name in transform_parameters:
        # pyre: param is declared to have type `float` but is used as
        # pyre-fixme[9]: type `Optional[typing.Union[bool, float, str]]`.
        param: float = parameters.get(p_name)
        rounded_parameters[p_name] = randomized_round(param)
    return rounded_parameters



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import enum
import json
import os
import pkgutil
import uuid
from typing import Dict

import plotly.offline as plotly_offline
from ax.plot.base import AxPlotConfig, AxPlotTypes
from jinja2 import Template

# Static plot module
PLOT_MODULE_NAME = "ax.plot"

# Rendering constants
DEFAULT_WIDTH = "100%"
DEFAULT_HEIGHT = 550
CSS_FILE = "ax/ax/plot/css/base.css"


# Common resources used in plotting (load with _load_js_resource)
class _AxPlotJSResources(enum.Enum):
    """Enum of common JS resources for plotting."""

    CSS_INJECTION = "css"
    HELPER_FXNS = "helpers"
    PLOTLY_OFFLINE = "plotly_offline"
    PLOTLY_ONLINE = "plotly_online"
    PLOTLY_REQUIRES = "plotly_requires"


# JS-based plots that are supported in Ax should be registered here
Ax_PLOT_REGISTRY: Dict[enum.Enum, str] = {AxPlotTypes.GENERIC: "generic_plotly.js"}


def _load_js_resource(resource_type: _AxPlotJSResources) -> str:
    """Convert plot config to corresponding JS code."""
    resource = pkgutil.get_data(
        PLOT_MODULE_NAME, os.path.join("js", "common", resource_type.value + ".js")
    )
    if resource is None:
        raise ValueError(f"Cannot find JS resource {resource_type.value}.")
    return resource.decode("utf8")


def _load_css_resource() -> str:
    resource = pkgutil.get_data(PLOT_MODULE_NAME, os.path.join("css", "base.css"))
    assert resource is not None
    return resource.decode("utf8")


def _js_requires(offline: bool = False) -> str:
    """Format JS requires for Plotly dependency.

    Args:
        offline: if True, inject entire Plotly library for offline use.

    Returns:
        str: <script> block with Plotly dependency.

    """
    helper_fxns = _load_js_resource(_AxPlotJSResources.HELPER_FXNS)
    if offline:
        script = Template(_load_js_resource(_AxPlotJSResources.PLOTLY_OFFLINE)).render(
            library=plotly_offline.offline.get_plotlyjs()
        )
    else:
        script = _load_js_resource(_AxPlotJSResources.PLOTLY_ONLINE)
    return script + helper_fxns


def _get_plot_js(
    config: AxPlotConfig,
    plot_module_name: str,
    plot_resources: Dict[enum.Enum, str],
    plotdivid: str,
) -> str:
    """Convert plot config to corresponding JS code."""
    if not isinstance(config, AxPlotConfig):
        raise ValueError("Config must be instance of AxPlotConfig.")
    js_template = pkgutil.get_data(
        plot_module_name, os.path.join("js", plot_resources[config.plot_type])
    )
    if js_template is None:
        raise ValueError(f"Cannot find JS template {plot_resources[config.plot_type]}.")
    return Template(js_template.decode("utf8")).render(
        id=json.dumps(plotdivid), **{k: json.dumps(v) for k, v in config.data.items()}
    )


def _wrap_js(script: str) -> str:
    """Wrap JS in <script></script> tag for injection into HTML."""
    return "<script type='text/javascript'>{script}</script>".format(script=script)


def _plot_js_to_html(js_script: str, plotdivid: str) -> str:
    """Embed JS script for Plotly plot in HTML.

    Result is a <div> block with a unique id set in `plotdivid` and a
    <script> block that contains the JS needed to render the plot inside the
    div.

    Args:
        js_script: JS for rendering plot.
        plotdivid: unique string ID for div.

    """
    plot_div = (
        '<div id="{id}" style="width: {width};" class="plotly-graph-div">' "</div>"
    ).format(id=plotdivid, width=DEFAULT_WIDTH)
    plot_js = Template(_load_js_resource(_AxPlotJSResources.PLOTLY_REQUIRES)).render(
        script=js_script
    )
    return plot_div + _wrap_js(plot_js)


def plot_config_to_html(
    plot_config: AxPlotConfig,
    plot_module_name: str = PLOT_MODULE_NAME,
    plot_resources: Dict[enum.Enum, str] = Ax_PLOT_REGISTRY,
    inject_helpers: bool = False,
) -> str:
    """Generate HTML + JS corresponding from a plot config."""
    plotdivid = uuid.uuid4().hex
    plot_js = _get_plot_js(plot_config, plot_module_name, plot_resources, plotdivid)
    if inject_helpers:
        helper_fxns = _load_js_resource(_AxPlotJSResources.HELPER_FXNS)
        plot_js = helper_fxns + plot_js
    return _plot_js_to_html(plot_js, plotdivid)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from unittest import mock
from unittest.mock import Mock

import numpy as np
from ax.core.metric import Metric
from ax.core.objective import Objective
from ax.core.observation import Observation, ObservationData, ObservationFeatures
from ax.core.optimization_config import OptimizationConfig
from ax.core.outcome_constraint import ComparisonOp, OutcomeConstraint
from ax.core.parameter import (
    ChoiceParameter,
    FixedParameter,
    ParameterType,
    RangeParameter,
)
from ax.core.search_space import SearchSpace
from ax.exceptions.core import UserInputError
from ax.modelbridge.discrete import _get_parameter_values, DiscreteModelBridge
from ax.models.discrete_base import DiscreteModel
from ax.utils.common.testutils import TestCase


class DiscreteModelBridgeTest(TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.parameters = [
            ChoiceParameter("x", ParameterType.FLOAT, values=[0, 1]),
            ChoiceParameter("y", ParameterType.STRING, values=["foo", "bar"]),
            FixedParameter("z", ParameterType.BOOL, value=True),
        ]
        parameter_constraints = []

        # pyre-fixme[6]: For 1st param expected `List[Parameter]` but got
        #  `List[Union[ChoiceParameter, FixedParameter]]`.
        self.search_space = SearchSpace(self.parameters, parameter_constraints)

        self.observation_features = [
            ObservationFeatures(parameters={"x": 0, "y": "foo", "z": True}),
            ObservationFeatures(parameters={"x": 1, "y": "foo", "z": True}),
            ObservationFeatures(parameters={"x": 1, "y": "bar", "z": True}),
        ]
        self.observation_data = [
            ObservationData(
                metric_names=["a", "b"],
                means=np.array([1.0, -1.0]),
                covariance=np.array([[1.0, 4.0], [4.0, 6.0]]),
            ),
            ObservationData(
                metric_names=["a", "b"],
                means=np.array([2.0, -2.0]),
                covariance=np.array([[2.0, 5.0], [5.0, 7.0]]),
            ),
            ObservationData(
                metric_names=["a"], means=np.array([3.0]), covariance=np.array([[3.0]])
            ),
        ]
        self.observations = [
            Observation(
                features=self.observation_features[i],
                data=self.observation_data[i],
                arm_name=str(i),
            )
            for i in range(3)
        ]
        self.pending_observations = {
            "b": [ObservationFeatures(parameters={"x": 0, "y": "foo", "z": True})]
        }
        self.model_gen_options = {"option": "yes"}

    @mock.patch(
        "ax.modelbridge.discrete.DiscreteModelBridge.__init__", return_value=None
    )
    def test_fit(self, mock_init: Mock) -> None:
        # pyre-fixme[20]: Argument `model` expected.
        ma = DiscreteModelBridge()
        ma._training_data = self.observations
        model = mock.create_autospec(DiscreteModel, instance=True)
        ma._fit(model, self.search_space, self.observations)
        self.assertEqual(ma.parameters, ["x", "y", "z"])
        self.assertEqual(sorted(ma.outcomes), ["a", "b"])
        Xs = {
            "a": [[0, "foo", True], [1, "foo", True], [1, "bar", True]],
            "b": [[0, "foo", True], [1, "foo", True]],
        }
        Ys = {"a": [[1.0], [2.0], [3.0]], "b": [[-1.0], [-2.0]]}
        Yvars = {"a": [[1.0], [2.0], [3.0]], "b": [[6.0], [7.0]]}
        parameter_values = [[0.0, 1.0], ["foo", "bar"], [True]]
        model_fit_args = model.fit.mock_calls[0][2]
        for i, x in enumerate(model_fit_args["Xs"]):
            self.assertEqual(x, Xs[ma.outcomes[i]])
        for i, y in enumerate(model_fit_args["Ys"]):
            self.assertEqual(y, Ys[ma.outcomes[i]])
        for i, v in enumerate(model_fit_args["Yvars"]):
            self.assertEqual(v, Yvars[ma.outcomes[i]])
        self.assertEqual(model_fit_args["parameter_values"], parameter_values)

        sq_obs = Observation(
            features=ObservationFeatures({}), data=self.observation_data[0]
        )
        with self.assertRaises(ValueError):
            ma._fit(model, self.search_space, self.observations + [sq_obs])

    @mock.patch(
        "ax.modelbridge.discrete.DiscreteModelBridge.__init__", return_value=None
    )
    def test_predict(self, mock_init: Mock) -> None:
        # pyre-fixme[20]: Argument `model` expected.
        ma = DiscreteModelBridge()
        model = mock.MagicMock(DiscreteModel, autospec=True, instance=True)
        model.predict.return_value = (
            np.array([[1.0, -1], [2.0, -2]]),
            np.stack(
                (np.array([[1.0, 4.0], [4.0, 6]]), np.array([[2.0, 5.0], [5.0, 7]]))
            ),
        )
        ma.model = model
        ma.parameters = ["x", "y", "z"]
        ma.outcomes = ["a", "b"]
        observation_data = ma._predict(self.observation_features)
        X = [[0, "foo", True], [1, "foo", True], [1, "bar", True]]
        self.assertTrue(model.predict.mock_calls[0][2]["X"], X)
        for i, od in enumerate(observation_data):
            self.assertEqual(od, self.observation_data[i])

    @mock.patch(
        "ax.modelbridge.discrete.DiscreteModelBridge.__init__", return_value=None
    )
    def test_gen(self, mock_init: Mock) -> None:
        # Test with constraints
        optimization_config = OptimizationConfig(
            objective=Objective(Metric("a"), minimize=True),
            outcome_constraints=[
                OutcomeConstraint(Metric("b"), ComparisonOp.GEQ, 2, False)
            ],
        )
        # pyre-fixme[20]: Argument `model` expected.
        ma = DiscreteModelBridge()
        # Test validation.
        with self.assertRaisesRegex(UserInputError, "positive integer or -1."):
            ma._validate_gen_inputs(n=0)
        ma._validate_gen_inputs(n=-1)
        # Test rest of gen.
        model = mock.MagicMock(DiscreteModel, autospec=True, instance=True)
        model.gen.return_value = ([[0.0, 2.0, 3.0], [1.0, 1.0, 3.0]], [1.0, 2.0], {})
        ma.model = model
        ma.parameters = ["x", "y", "z"]
        ma.outcomes = ["a", "b"]
        gen_results = ma._gen(
            n=3,
            search_space=self.search_space,
            optimization_config=optimization_config,
            pending_observations=self.pending_observations,
            fixed_features=ObservationFeatures({}),
            # pyre-fixme[6]: For 6th param expected `Optional[Dict[str, Union[None,
            #  Dict[str, typing.Any], OptimizationConfig, AcquisitionFunction, float,
            #  int, str]]]` but got `Dict[str, str]`.
            model_gen_options=self.model_gen_options,
        )
        gen_args = model.gen.mock_calls[0][2]
        self.assertEqual(gen_args["n"], 3)
        self.assertEqual(
            gen_args["parameter_values"], [[0.0, 1.0], ["foo", "bar"], [True]]
        )
        self.assertTrue(
            np.array_equal(gen_args["objective_weights"], np.array([-1.0, 0.0]))
        )
        self.assertTrue(
            np.array_equal(gen_args["outcome_constraints"][0], np.array([[0.0, -1.0]]))
        )
        self.assertTrue(
            np.array_equal(gen_args["outcome_constraints"][1], np.array([[-2]]))
        )
        self.assertEqual(gen_args["pending_observations"][0], [])
        self.assertEqual(gen_args["pending_observations"][1], [[0, "foo", True]])
        self.assertEqual(gen_args["model_gen_options"], {"option": "yes"})
        self.assertEqual(
            gen_results.observation_features[0].parameters,
            {"x": 0.0, "y": 2.0, "z": 3.0},
        )
        self.assertEqual(
            gen_results.observation_features[1].parameters,
            {"x": 1.0, "y": 1.0, "z": 3.0},
        )
        self.assertEqual(gen_results.weights, [1.0, 2.0])

        # Test with no constraints, no fixed feature, no pending observations
        # pyre-fixme[6]: For 1st param expected `List[Parameter]` but got
        #  `List[Union[ChoiceParameter, FixedParameter]]`.
        search_space = SearchSpace(self.parameters[:2])
        optimization_config.outcome_constraints = []
        ma.parameters = ["x", "y"]
        ma._gen(
            n=3,
            search_space=search_space,
            optimization_config=optimization_config,
            pending_observations={},
            fixed_features=ObservationFeatures({}),
            model_gen_options={},
        )
        gen_args = model.gen.mock_calls[1][2]
        self.assertEqual(gen_args["parameter_values"], [[0.0, 1.0], ["foo", "bar"]])
        self.assertIsNone(gen_args["outcome_constraints"])
        self.assertIsNone(gen_args["pending_observations"])

        # Test validation
        optimization_config = OptimizationConfig(
            objective=Objective(Metric("a"), minimize=False),
            outcome_constraints=[
                OutcomeConstraint(Metric("b"), ComparisonOp.GEQ, 2, True)
            ],
        )
        with self.assertRaises(ValueError):
            ma._gen(
                n=3,
                search_space=search_space,
                optimization_config=optimization_config,
                pending_observations={},
                fixed_features=ObservationFeatures({}),
                model_gen_options={},
            )

    @mock.patch(
        "ax.modelbridge.discrete.DiscreteModelBridge.__init__", return_value=None
    )
    def test_cross_validate(self, mock_init: Mock) -> None:
        # pyre-fixme[20]: Argument `model` expected.
        ma = DiscreteModelBridge()
        model = mock.MagicMock(DiscreteModel, autospec=True, instance=True)
        model.cross_validate.return_value = (
            np.array([[1.0, -1], [2.0, -2]]),
            np.stack(
                (np.array([[1.0, 4.0], [4.0, 6]]), np.array([[2.0, 5.0], [5.0, 7]]))
            ),
        )
        ma.model = model
        ma.parameters = ["x", "y", "z"]
        ma.outcomes = ["a", "b"]
        observation_data = ma._cross_validate(
            search_space=self.search_space,
            cv_training_data=self.observations,
            cv_test_points=self.observation_features,
        )
        Xs = [
            [[0, "foo", True], [1, "foo", True], [1, "bar", True]],
            [[0, "foo", True], [1, "foo", True]],
        ]
        Ys = [[[1.0], [2.0], [3.0]], [[-1.0], [-2.0]]]
        Yvars = [[[1.0], [2.0], [3.0]], [[6.0], [7.0]]]
        Xtest = [[0, "foo", True], [1, "foo", True], [1, "bar", True]]
        # Transform to arrays:
        model_cv_args = model.cross_validate.mock_calls[0][2]
        for i, x in enumerate(model_cv_args["Xs_train"]):
            self.assertEqual(x, Xs[i])
        for i, y in enumerate(model_cv_args["Ys_train"]):
            self.assertEqual(y, Ys[i])
        for i, v in enumerate(model_cv_args["Yvars_train"]):
            self.assertEqual(v, Yvars[i])
        self.assertEqual(model_cv_args["X_test"], Xtest)
        # Transform from arrays:
        for i, od in enumerate(observation_data):
            self.assertEqual(od, self.observation_data[i])

    def test_get_parameter_values(self) -> None:
        parameter_values = _get_parameter_values(self.search_space, ["x", "y", "z"])
        self.assertEqual(parameter_values, [[0.0, 1.0], ["foo", "bar"], [True]])
        # pyre-fixme[6]: For 1st param expected `List[Parameter]` but got
        #  `List[Union[ChoiceParameter, FixedParameter]]`.
        search_space = SearchSpace(self.parameters)
        search_space._parameters["x"] = RangeParameter(
            "x", ParameterType.FLOAT, 0.1, 0.4
        )
        with self.assertRaises(ValueError):
            _get_parameter_values(search_space, ["x", "y", "z"])



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import math
from time import sleep
from unittest import mock
from unittest.mock import patch, PropertyMock

import numpy as np
from ax.core.arm import Arm
from ax.core.base_trial import TrialStatus
from ax.core.batch_trial import BatchTrial, GeneratorRunStruct
from ax.core.experiment import Experiment
from ax.core.generator_run import GeneratorRun, GeneratorRunType
from ax.core.parameter import FixedParameter, ParameterType
from ax.core.search_space import SearchSpace
from ax.exceptions.core import UnsupportedError
from ax.runners.synthetic import SyntheticRunner
from ax.utils.common.testutils import TestCase
from ax.utils.common.typeutils import checked_cast
from ax.utils.testing.core_stubs import (
    get_abandoned_arm,
    get_arm,
    get_arm_weights1,
    get_arms,
    get_experiment,
    get_generator_run,
    get_generator_run2,
    get_weights,
)


class BatchTrialTest(TestCase):
    def setUp(self) -> None:
        self.experiment = get_experiment()
        self.experiment.status_quo = None
        self.batch = self.experiment.new_batch_trial()
        arms = get_arms()
        weights = get_weights()
        self.status_quo = arms[0]
        self.sq_weight = weights[0]
        self.arms = arms[1:]
        self.weights = weights[1:]
        self.batch.add_arms_and_weights(arms=self.arms, weights=self.weights)

    def test__validate_can_attach_data(self) -> None:
        self.batch.mark_running(no_runner_required=True)
        self.batch.mark_completed()

        expected_msg = (
            "Trial 0 already has status 'COMPLETED', so data cannot be attached."
        )
        with self.assertRaisesRegex(UnsupportedError, expected_msg):
            self.batch._validate_can_attach_data()

    def test_Eq(self) -> None:
        new_batch_trial = self.experiment.new_batch_trial()
        self.assertNotEqual(self.batch, new_batch_trial)

        abandoned_arm = get_abandoned_arm()
        abandoned_arm_2 = get_abandoned_arm()
        self.assertEqual(abandoned_arm, abandoned_arm_2)

    def test_BasicProperties(self) -> None:
        self.assertEqual(self.experiment, self.batch.experiment)
        self.assertEqual(self.batch.index, 0)
        self.assertEqual(self.batch.status, TrialStatus.CANDIDATE)
        self.assertIsNotNone(self.batch.time_created)
        self.assertEqual(
            self.batch.generator_run_structs[0].generator_run.generator_run_type,
            GeneratorRunType.MANUAL.name,
        )

        # Test empty arms
        self.assertEqual(len(self.experiment.new_batch_trial().abandoned_arms), 0)

    def test_UndefinedSetters(self) -> None:
        with self.assertRaises(NotImplementedError):
            self.batch.arm_weights = get_arm_weights1()

        with self.assertRaises(NotImplementedError):
            self.batch.status = TrialStatus.RUNNING

    def test_BasicSetter(self) -> None:
        self.batch.runner = SyntheticRunner()
        self.assertIsNotNone(self.batch.runner)

        self.batch.trial_type = None
        self.assertIsNone(self.batch.trial_type)

        # Default experiment only supports None as trial_type
        with self.assertRaises(ValueError):
            self.batch.trial_type = ""

    def test_AddArm(self) -> None:
        self.assertEqual(len(self.batch.arms), len(self.arms))
        self.assertEqual(len(self.batch.generator_run_structs), 1)
        self.assertEqual(sum(self.batch.weights), sum(self.weights))

        arm_parameters = get_arm().parameters
        arm_parameters["w"] = 5.0
        self.batch.add_arm(Arm(arm_parameters), 3)

        self.assertEqual(self.batch.arms_by_name["0_2"], self.batch.arms[2])
        self.assertEqual(len(self.batch.arms), len(self.arms) + 1)
        self.assertEqual(len(self.batch.generator_run_structs), 2)
        self.assertEqual(sum(self.batch.weights), sum(self.weights) + 3)

    def test_add_generator_run(self) -> None:
        self.assertEqual(len(self.batch.arms), len(self.arms))
        self.assertEqual(len(self.batch.generator_run_structs), 1)
        self.assertEqual(sum(self.batch.weights), sum(self.weights))

        # Overwrite the GS index to not-None.
        self.batch._generation_step_index = 0

        # one of these arms already exists on the BatchTrial,
        # so we should just update its weight
        new_arms = [
            Arm(parameters={"w": 0.75, "x": 1, "y": "foo", "z": True}),
            Arm(parameters={"w": 1.4, "x": 5, "y": "bar", "z": False}),
        ]
        new_weights = [0.75, 0.25]
        gr = GeneratorRun(arms=new_arms, weights=new_weights)
        self.batch.add_generator_run(gr, 2.0)

        self.assertEqual(len(self.batch.arms), len(self.arms) + 1)
        self.assertEqual(len(self.batch.generator_run_structs), 2)
        self.assertEqual(sum(self.batch.weights), sum(self.weights) + 2)
        # Check the GS index was not overwritten to None.
        self.assertEqual(self.batch._generation_step_index, 0)

    def test_InitWithGeneratorRun(self) -> None:
        generator_run = GeneratorRun(arms=self.arms, weights=self.weights)
        batch = self.experiment.new_batch_trial(generator_run=generator_run)
        batch.add_arms_and_weights(arms=self.arms, weights=self.weights)
        self.assertEqual(self.batch.arms_by_name["0_0"], self.batch.arms[0])
        self.assertEqual(self.batch.arms_by_name["0_1"], self.batch.arms[1])
        self.assertEqual(len(batch.arms), len(self.arms))
        self.assertEqual(len(self.batch.generator_run_structs), 1)

    def test_StatusQuoOverlap(self) -> None:
        new_sq = Arm(parameters={"w": 0.95, "x": 1, "y": "foo", "z": True})
        # Set status quo to existing arm
        self.batch.set_status_quo_with_weight(self.arms[0], self.sq_weight)
        # Status quo weight is set to the average of other arms' weights.
        # In this case, there are only two arms: 0_0 (SQ) and 0_1 (not SQ).
        # So their weights are equal, as weight(0_0) = avg(weight(0_1)).
        self.assertEqual(self.batch.weights[0], self.batch.weights[1])
        self.assertTrue(self.batch.status_quo.parameters == self.arms[0].parameters)
        self.assertEqual(self.batch.status_quo.name, self.batch.arms[0].name)
        self.assertEqual(self.batch.arm_weights[self.batch.arms[0]], self.sq_weight)
        self.assertEqual(sum(self.batch.weights), self.weights[1] + self.sq_weight)

        # Set status quo to new arm, add it
        self.batch.set_status_quo_with_weight(new_sq, self.sq_weight)
        self.assertEqual(self.batch.status_quo.name, "status_quo_0")
        self.batch.add_arms_and_weights([new_sq])
        self.assertEqual(
            self.batch.generator_run_structs[1].generator_run.arms[0].name,
            "status_quo_0",
        )

    def test_StatusQuo(self) -> None:
        tot_weight = sum(self.batch.weights)
        new_sq = Arm(parameters={"w": 0.95, "x": 1, "y": "foo", "z": True})

        # Test negative weight
        with self.assertRaises(ValueError):
            self.batch.set_status_quo_with_weight(new_sq, -1)

        # Test that directly setting the status quo raises an error
        with self.assertRaises(NotImplementedError):
            self.batch.status_quo = new_sq

        # Set status quo to new arm
        self.batch.set_status_quo_with_weight(new_sq, self.sq_weight)
        self.assertTrue(self.batch.status_quo == new_sq)
        self.assertEqual(self.batch.status_quo.name, "status_quo_0")
        self.assertEqual(sum(self.batch.weights), tot_weight + self.sq_weight)
        # sq weight should be ignored when sq is None
        self.batch.unset_status_quo()
        self.assertEqual(sum(self.batch.weights), tot_weight)

        # Verify experiment status quo gets set on init
        self.experiment.status_quo = self.status_quo
        batch2 = self.batch.clone()
        self.assertEqual(batch2.status_quo, self.experiment.status_quo)

        # Since optimize_for_power was not set, the weight override should not be
        # And the status quo shoudl not appear in arm_weights
        self.assertIsNone(batch2._status_quo_weight_override)
        self.assertTrue(batch2.status_quo not in batch2.arm_weights)
        self.assertEqual(sum(batch2.weights), sum(self.weights))

        # Try setting sq to existing arm with different name
        with self.assertRaises(ValueError):
            self.batch.set_status_quo_with_weight(
                Arm(new_sq.parameters, name="new_name"), 1
            )

    def test_StatusQuoOptimizeForPower(self) -> None:
        self.experiment.status_quo = self.status_quo
        batch = self.experiment.new_batch_trial(optimize_for_power=True)
        self.assertEqual(batch._status_quo_weight_override, 1)

        self.experiment.status_quo = None
        with self.assertRaises(ValueError):
            batch = self.experiment.new_batch_trial(optimize_for_power=True)

        batch.add_arms_and_weights(arms=[])
        self.assertTrue(batch._status_quo_weight_override, 1)

        batch.add_arms_and_weights(arms=self.arms, weights=self.weights)
        expected_status_quo_weight = math.sqrt(sum(self.weights))
        self.assertTrue(
            math.isclose(batch._status_quo_weight_override, expected_status_quo_weight)
        )
        self.assertTrue(
            math.isclose(
                batch.arm_weights[batch.status_quo], expected_status_quo_weight
            )
        )

    def test_ArmsByName(self) -> None:
        # Initializes empty
        newbatch = self.experiment.new_batch_trial()
        self.assertEqual(newbatch.arms_by_name, {})

        # Correct when arms have been added
        self.assertEqual(self.batch.arms_by_name["0_0"], self.batch.arms[0])
        self.assertEqual(self.batch.arms_by_name["0_1"], self.batch.arms[1])

        # Refreshed when arm is added
        newbatch.add_arm(self.batch.arms[0])
        self.assertEqual(newbatch.arms_by_name, {"0_0": self.batch.arms[0]})

        # Refreshed when status quo is set
        newbatch.set_status_quo_with_weight(self.batch.arms[1], 1.0)
        self.assertEqual(
            newbatch.arms_by_name,
            {"0_0": self.batch.arms[0], "0_1": self.batch.arms[1]},
        )

        # Refreshed when status quo is unset
        newbatch.unset_status_quo()
        self.assertEqual(
            newbatch.arms_by_name,
            {"0_0": self.batch.arms[0]},
        )

    def test_BatchLifecycle(self) -> None:
        # Check that state of trial statuses mapping on experiment: there should only be
        # one index, 0, among the `CANDIDATE` trials.
        trial_idcs_by_status = iter(self.experiment.trial_indices_by_status.values())
        self.assertEqual(next(trial_idcs_by_status), {0})  # `CANDIDATE` trial indices
        # ALl other trial statuses should not yet have trials carry them.
        self.assertTrue(all(len(idcs) == 0 for idcs in trial_idcs_by_status))
        staging_mock = PropertyMock()
        with patch.object(SyntheticRunner, "staging_required", staging_mock):
            mock_runner = SyntheticRunner()
            staging_mock.return_value = True
            self.batch.runner = mock_runner
            self.batch.run()
            self.assertEqual(self.batch.status, TrialStatus.STAGED)
            # Check that the trial statuses mapping on experiment has been updated.
            self.assertEqual(
                self.experiment.trial_indices_by_status[TrialStatus.STAGED], {0}
            )
            self.assertTrue(
                # pyre-fixme[6]: For 1st param expected `Iterable[object]` but got
                #  `bool`.
                all(len(idcs) == 0)
                for status, idcs in self.experiment.trial_indices_by_status.items()
                if status != TrialStatus.STAGED
            )
            self.assertIsNotNone(self.batch.time_staged)
            self.assertTrue(self.batch.status.is_deployed)
            self.assertFalse(self.batch.status.expecting_data)

            # Cannot change arms or runner once run
            with self.assertRaises(ValueError):
                self.batch.add_arms_and_weights(arms=self.arms, weights=self.weights)

            with self.assertRaises(ValueError):
                self.batch.runner = None

            # Cannot run batch that was already run
            with self.assertRaises(ValueError):
                self.batch.run()

            self.batch.mark_running()
            self.assertEqual(self.batch.status, TrialStatus.RUNNING)
            # Check that the trial statuses mapping on experiment has been updated.
            self.assertEqual(
                self.experiment.trial_indices_by_status[TrialStatus.RUNNING], {0}
            )
            self.assertTrue(
                # pyre-fixme[6]: For 1st param expected `Iterable[object]` but got
                #  `bool`.
                all(len(idcs) == 0)
                for status, idcs in self.experiment.trial_indices_by_status.items()
                if status != TrialStatus.RUNNING
            )
            self.assertIsNotNone(self.batch.time_run_started)
            self.assertTrue(self.batch.status.expecting_data)

            self.batch.complete()
            # Cannot complete that which is already completed
            with self.assertRaises(ValueError):
                self.batch.complete()

            # Verify trial is completed
            self.assertEqual(self.batch.status, TrialStatus.COMPLETED)
            # Check that the trial statuses mapping on experiment has been updated.
            self.assertEqual(
                self.experiment.trial_indices_by_status[TrialStatus.COMPLETED], {0}
            )
            self.assertTrue(
                # pyre-fixme[6]: For 1st param expected `Iterable[object]` but got
                #  `bool`.
                all(len(idcs) == 0)
                for status, idcs in self.experiment.trial_indices_by_status.items()
                if status != TrialStatus.COMPLETED
            )
            self.assertIsNotNone(self.batch.time_completed)
            self.assertTrue(self.batch.status.is_terminal)

            # Cannot change status after BatchTrial is completed
            with self.assertRaises(ValueError):
                self.batch.mark_staged()

            with self.assertRaises(ValueError):
                self.batch.mark_completed()

            with self.assertRaises(ValueError):
                self.batch.mark_running()

            with self.assertRaises(ValueError):
                self.batch.mark_abandoned()

            with self.assertRaises(ValueError):
                self.batch.mark_failed()

            # Check that the trial statuses mapping on experiment is updated when
            # trial status is set hackily / directly, without using `mark_X`.
            self.batch._status = TrialStatus.CANDIDATE
            self.assertEqual(
                self.experiment.trial_indices_by_status[TrialStatus.CANDIDATE], {0}
            )
            self.assertTrue(
                # pyre-fixme[6]: For 1st param expected `Iterable[object]` but got
                #  `bool`.
                all(len(idcs) == 0)
                for status, idcs in self.experiment.trial_indices_by_status.items()
                if status != TrialStatus.CANDIDATE
            )

    def test_AbandonBatchTrial(self) -> None:
        reason = "BatchTrial behaved poorly"
        self.batch.mark_abandoned(reason)

        self.assertEqual(self.batch.status, TrialStatus.ABANDONED)
        self.assertIsNotNone(self.batch.time_completed)
        self.assertEqual(self.batch.abandoned_reason, reason)

    def test_FailedBatchTrial(self) -> None:
        self.batch.runner = SyntheticRunner()
        self.batch.run()
        self.batch.mark_failed()

        self.assertEqual(self.batch.status, TrialStatus.FAILED)
        self.assertIsNotNone(self.batch.time_completed)

    def test_EarlyStoppedBatchTrial(self) -> None:
        self.batch.runner = SyntheticRunner()
        self.batch.run()
        self.batch.mark_early_stopped()

        self.assertEqual(self.batch.status, TrialStatus.EARLY_STOPPED)
        self.assertIsNotNone(self.batch.time_completed)

    def test_AbandonArm(self) -> None:
        arm = self.batch.arms[0]
        reason = "Bad arm"
        self.batch.mark_arm_abandoned(arm.name, reason)
        self.assertEqual(len(self.batch.abandoned_arms), 1)
        self.assertEqual(self.batch.abandoned_arms[0], arm)

        self.assertEqual(len(self.batch.abandoned_arms_metadata), 1)
        metadata = self.batch.abandoned_arms_metadata[0]
        self.assertEqual(metadata.reason, reason)
        self.assertEqual(metadata.name, arm.name)

        # Fail to abandon arm not in BatchTrial
        with self.assertRaises(ValueError):
            self.batch.mark_arm_abandoned(
                Arm(parameters={"x": 3, "y": "fooz", "z": False})
            )

    @mock.patch(
        f"{get_experiment.__module__}.Experiment.supports_trial_type",
        return_value=True,
    )
    def test_clone_to(self, _) -> None:
        experiment = get_experiment()
        experiment.status_quo = None
        batch = experiment.new_batch_trial(ttl_seconds=123, trial_type="foo")
        arms = get_arms()
        weights = get_weights()
        status_quo = arms[0]
        arms = arms[1:]
        weights = weights[1:]
        batch.add_arms_and_weights(arms=arms, weights=weights)
        batch.update_run_metadata(metadata={"foo": "bar"})
        batch.update_stop_metadata(metadata={"bar": "baz"})

        # cloning the trial and attached it to the same experiment
        new_batch_trial_0 = batch.clone_to()
        # cloning the trial and attached it to a new experiment
        new_experiment = get_experiment()
        new_experiment.status_quo = None
        batch.clone_to(new_experiment)
        new_batch_trial_1 = checked_cast(BatchTrial, new_experiment.trials[0])

        self.assertEqual(new_batch_trial_0.index, 1)
        # Set index to original trial's value for equality check.
        new_batch_trial_0._index = batch.index
        new_batch_trial_0._time_created = batch._time_created
        new_batch_trial_1._time_created = batch._time_created
        self.assertEqual(new_batch_trial_0, batch)
        self.assertEqual(new_batch_trial_1, batch)

        # make sure modifying the cloned batch trial does not affect original one
        new_batch_trial_1.add_arm(
            Arm(name="new_arm", parameters={"w": 2.6, "x": 2, "y": "baz", "z": True})
        )
        self.assertEqual(len(new_batch_trial_1.arms), 3)
        self.assertEqual(len(batch.arms), 2)

        # cloning a trial that has status quo arm
        status_quo = Arm(
            name="status_quo", parameters={"w": 0.0, "x": 1, "y": "foo", "z": True}
        )
        batch.set_status_quo_and_optimize_power(status_quo)
        batch.mark_running(no_runner_required=True)
        new_batch_trial = batch.clone_to()
        self.assertEqual(new_batch_trial.index, 2)
        # Set index & time_created to original trial's value for equality check.
        new_batch_trial._index = batch.index
        new_batch_trial._time_created = batch._time_created
        self.assertEqual(new_batch_trial, batch)

    def test_Runner(self) -> None:
        # Verify BatchTrial without runner will fail
        with self.assertRaises(ValueError):
            self.batch.run()

        # Verify mark running without runner will fail
        with self.assertRaises(ValueError):
            self.batch.mark_running()

        self.batch.runner = SyntheticRunner()
        self.batch.run()
        self.assertEqual(self.batch.deployed_name, "test_0")
        self.assertNotEqual(len(self.batch.run_metadata.keys()), 0)
        self.assertEqual(self.batch.status, TrialStatus.RUNNING)

        # Verify setting runner on experiment but not BatchTrial
        # Also mock staging_required to be false
        staging_mock = PropertyMock()
        with patch.object(SyntheticRunner, "staging_required", staging_mock):
            mock_runner = SyntheticRunner()
            staging_mock.return_value = True

            self.experiment.runner = mock_runner
            b2 = self.experiment.new_batch_trial()
            b2.run()
            self.assertEqual(b2.deployed_name, "test_1")
            self.assertEqual(b2.status, TrialStatus.STAGED)

    def test_IsFactorial(self) -> None:
        self.assertFalse(self.batch.is_factorial)

        # Insufficient factors
        small_experiment = Experiment(
            name="small_test",
            search_space=SearchSpace([FixedParameter("a", ParameterType.INT, 4)]),
        )
        small_trial = small_experiment.new_batch_trial().add_arm(Arm({"a": 4}))
        self.assertFalse(small_trial.is_factorial)

        new_batch_trial = self.experiment.new_batch_trial()
        new_batch_trial.add_arms_and_weights(
            arms=[
                Arm(parameters={"w": 0.75, "x": 1, "y": "foo", "z": True}),
                Arm(parameters={"w": 0.75, "x": 2, "y": "foo", "z": True}),
                Arm(parameters={"w": 0.77, "x": 1, "y": "foo", "z": True}),
            ]
        )
        self.assertFalse(new_batch_trial.is_factorial)

        new_batch_trial = self.experiment.new_batch_trial()
        new_batch_trial.add_arms_and_weights(
            arms=[
                Arm(parameters={"w": 0.77, "x": 1, "y": "foo", "z": True}),
                Arm(parameters={"w": 0.77, "x": 2, "y": "foo", "z": True}),
                Arm(parameters={"w": 0.75, "x": 1, "y": "foo", "z": True}),
                Arm(parameters={"w": 0.75, "x": 2, "y": "foo", "z": True}),
            ]
        )
        self.assertTrue(new_batch_trial.is_factorial)

    def test_NormalizedArmWeights(self) -> None:
        new_batch_trial = self.experiment.new_batch_trial()
        parameterizations = [
            {"w": 0.75, "x": 1, "y": "foo", "z": True},
            {"w": 0.77, "x": 2, "y": "foo", "z": True},
        ]
        # pyre-fixme[6]: For 1st param expected `Dict[str, Union[None, bool, float,
        #  int, str]]` but got `Dict[str, Union[float, str]]`.
        arms = [Arm(parameters=p) for i, p in enumerate(parameterizations)]
        new_batch_trial.add_arms_and_weights(arms=arms, weights=[2, 1])

        # test normalizing to 1
        arm_weights = new_batch_trial.normalized_arm_weights()
        # self.assertEqual(list(arm_weights.keys()), arms)
        batch_arm_parameters = [arm.parameters for arm in list(arm_weights.keys())]
        arm_parameters = [arm.parameters for arm in arms]
        self.assertEqual(batch_arm_parameters, arm_parameters)
        self.assertTrue(np.allclose(list(arm_weights.values()), [2 / 3, 1 / 3]))

        # test normalizing to 100
        arm_weights = new_batch_trial.normalized_arm_weights(total=100)
        batch_arm_parameters = [arm.parameters for arm in list(arm_weights.keys())]
        arm_parameters = [arm.parameters for arm in arms]
        self.assertEqual(batch_arm_parameters, arm_parameters)
        self.assertTrue(np.allclose(list(arm_weights.values()), [200 / 3, 100 / 3]))

        # test normalizing with truncation
        arm_weights = new_batch_trial.normalized_arm_weights(total=1, trunc_digits=2)
        batch_arm_parameters = [arm.parameters for arm in list(arm_weights.keys())]
        arm_parameters = [arm.parameters for arm in arms]
        self.assertEqual(batch_arm_parameters, arm_parameters)
        self.assertTrue(np.allclose(list(arm_weights.values()), [0.67, 0.33]))

    def test_AddGeneratorRunValidation(self) -> None:
        new_batch_trial = self.experiment.new_batch_trial()
        new_arms = [
            Arm(name="0_1", parameters={"w": 0.75, "x": 1, "y": "foo", "z": True}),
            Arm(name="0_2", parameters={"w": 0.75, "x": 1, "y": "foo", "z": True}),
        ]
        gr = GeneratorRun(arms=new_arms)
        with self.assertRaises(ValueError):
            new_batch_trial.add_generator_run(gr)

    def test_SetStatusQuoAndOptimizePower(self) -> None:
        batch_trial = self.experiment.new_batch_trial()
        status_quo = Arm(
            name="status_quo", parameters={"w": 0.0, "x": 1, "y": "foo", "z": True}
        )

        # Test adding status quo and optimizing power on empty batch
        batch_trial.set_status_quo_and_optimize_power(status_quo)
        self.assertEqual(batch_trial.arm_weights[status_quo], 1.0)

        # Test adding status quo and optimizing power on non-empty batch
        batch_trial = self.experiment.new_batch_trial()
        parameterizations = [
            {"w": 0.75, "x": 1, "y": "foo", "z": True},
            {"w": 0.77, "x": 2, "y": "foo", "z": True},
        ]
        # pyre-fixme[6]: For 1st param expected `Dict[str, Union[None, bool, float,
        #  int, str]]` but got `Dict[str, Union[float, str]]`.
        arms = [Arm(parameters=p) for i, p in enumerate(parameterizations)]
        batch_trial.add_arms_and_weights(arms=arms)
        batch_trial.set_status_quo_and_optimize_power(status_quo)
        self.assertEqual(batch_trial.arm_weights[status_quo], np.sqrt(2))

        # Test adding status quo and optimizing power when trial already
        # has a status quo
        batch_trial = self.experiment.new_batch_trial()
        batch_trial.set_status_quo_with_weight(status_quo, 1)
        self.assertEqual(batch_trial.arm_weights[status_quo], 1.0)
        batch_trial.add_arms_and_weights(arms=arms)
        batch_trial.set_status_quo_and_optimize_power(status_quo)
        self.assertEqual(batch_trial.arm_weights[status_quo], np.sqrt(2))
        # Since status quo is not in the generator runs, all of its weight
        # comes from _status_quo_weight_override
        self.assertEqual(batch_trial._status_quo_weight_override, np.sqrt(2))

        # Test adding status quo and optimizing power when status quo
        # is in the generator runs
        batch_trial = self.experiment.new_batch_trial()
        parameterizations = [
            {"w": 0.75, "x": 1, "y": "foo", "z": True},
            {"w": 0.77, "x": 2, "y": "foo", "z": True},
            {"w": 0.0, "x": 1, "y": "foo", "z": True},
        ]
        # pyre-fixme[6]: For 1st param expected `Dict[str, Union[None, bool, float,
        #  int, str]]` but got `Dict[str, Union[float, str]]`.
        arms = [Arm(parameters=p) for i, p in enumerate(parameterizations)]
        batch_trial.add_arms_and_weights(arms=arms)
        batch_trial.set_status_quo_and_optimize_power(status_quo)
        self.assertEqual(batch_trial.arm_weights[status_quo], np.sqrt(2))
        # Since status quo has a weight of 1 in the generator runs, only part of
        # its weight comes from _status_quo_weight_override
        self.assertEqual(batch_trial._status_quo_weight_override, np.sqrt(2))

    def test_Repr(self) -> None:
        self.assertEqual(
            str(self.batch),
            "BatchTrial(experiment_name='test', index=0, status=TrialStatus.CANDIDATE)",
        )

    def test_TTL(self) -> None:
        # Verify that TLL is checked on execution of the `status` property.
        self.batch.ttl_seconds = 1
        self.batch.mark_running(no_runner_required=True)
        self.assertTrue(self.batch.status.is_running)
        sleep(1)  # Wait 1 second for trial TTL to elapse.
        self.assertTrue(self.batch.status.is_failed)
        self.assertIn(0, self.experiment.trial_indices_by_status[TrialStatus.FAILED])

        # Verify that TTL is checked on `experiment.trial_indices_by_status`.
        batch_trial = self.experiment.new_batch_trial(ttl_seconds=1)
        batch_trial.mark_running(no_runner_required=True)
        self.assertTrue(batch_trial.status.is_running)
        sleep(1)  # Wait 1 second for trial TTL to elapse.
        self.assertIn(1, self.experiment.trial_indices_by_status[TrialStatus.FAILED])
        self.assertTrue(self.experiment.trials[1].status.is_failed)

        # Verify that TTL is checked on `experiment.trials`.
        batch_trial = self.experiment.new_batch_trial(ttl_seconds=1)
        batch_trial.mark_running(no_runner_required=True)
        self.assertTrue(batch_trial.status.is_running)
        self.assertIn(2, self.experiment.trial_indices_by_status[TrialStatus.RUNNING])
        sleep(1)  # Wait 1 second for trial TTL to elapse.
        self.experiment.trials
        # Check `_status`, not `status`, to ensure it's within `trials` that the status
        # was actually changed, not in `status`.
        self.assertEqual(batch_trial._status, TrialStatus.FAILED)
        self.assertIn(2, self.experiment.trial_indices_by_status[TrialStatus.FAILED])

    def test_get_candidate_metadata_from_all_generator_runs(self) -> None:
        gr_1 = get_generator_run()
        gr_2 = get_generator_run2()
        self.batch.add_generator_run(gr_1)
        # Arms are named when adding GR to trial, so reassign to have a GR that has
        # names arms.
        gr_1 = self.batch._generator_run_structs[-1].generator_run
        self.batch.add_generator_run(gr_2)
        gr_2 = self.batch._generator_run_structs[-1].generator_run
        # gr_2 has no candidate metadata; all candidate metadata should come from gr_1
        cand_metadata_expected = {
            a.name: gr_1.candidate_metadata_by_arm_signature[a.signature]
            for a in gr_1.arms
        }
        self.assertEqual(
            self.batch._get_candidate_metadata_from_all_generator_runs(),
            cand_metadata_expected,
        )
        for arm in self.batch.arms:
            self.assertEqual(
                cand_metadata_expected[arm.name],
                self.batch._get_candidate_metadata(arm.name),
            )
        self.assertRaises(
            ValueError, self.batch._get_candidate_metadata, "this_is_not_an_arm"
        )

        # Check that if we add cand. metadata to gr_2, it will appear in cand.
        # metadata for the batch.
        gr_3 = get_generator_run2()
        new_cand_metadata = {
            a.signature: {"md_key": f"md_val_{a.signature}"} for a in gr_3.arms
        }
        # pyre-fixme[8]: Attribute has type `Optional[Dict[str, Optional[Dict[str,
        #  typing.Any]]]]`; used as `Dict[str, Dict[str, str]]`.
        gr_3._candidate_metadata_by_arm_signature = new_cand_metadata
        self.batch.add_generator_run(gr_3)
        gr_3 = self.batch._generator_run_structs[-1].generator_run
        cand_metadata_expected.update(
            {
                a.name: gr_1.candidate_metadata_by_arm_signature[a.signature]
                for a in gr_1.arms
            }
        )
        self.assertEqual(
            self.batch._get_candidate_metadata_from_all_generator_runs(),
            cand_metadata_expected,
        )
        for arm in self.batch.arms:
            self.assertEqual(
                cand_metadata_expected[arm.name],
                self.batch._get_candidate_metadata(arm.name),
            )

    def test_Sortable(self) -> None:
        new_batch_trial = self.experiment.new_batch_trial()
        self.assertTrue(self.batch < new_batch_trial)

        abandoned_arm = get_abandoned_arm()
        abandoned_arm_2 = get_abandoned_arm()
        abandoned_arm_2.name = "0_1"
        self.assertTrue(abandoned_arm < abandoned_arm_2)

        generator_run = get_generator_run()
        generator_run_struct = GeneratorRunStruct(
            generator_run=generator_run, weight=1.0
        )
        generator_run_struct_2 = GeneratorRunStruct(
            generator_run=generator_run, weight=2.0
        )
        self.assertTrue(generator_run_struct < generator_run_struct_2)

    def test_attach_batch_trial_data(self) -> None:
        # Verify components before we attach trial data
        self.assertEqual(2, len(self.batch.arms))
        arm1_name = self.batch.arms[0].name
        arm2_name = self.batch.arms[1].name

        self.assertEqual(
            2,
            len(self.batch.experiment.metrics)
            - len(self.batch.experiment.tracking_metrics),
        )
        self.assertTrue("m1" in self.batch.experiment.metrics)
        self.assertTrue("m2" in self.batch.experiment.metrics)

        data = self.batch.lookup_data().df.to_dict(orient="index")
        self.assertTrue(len(data) == 0)

        # Attach data
        self.batch.attach_batch_trial_data(
            raw_data={
                arm1_name: {"m1": 1.0, "m2": 2.0},
                arm2_name: {"m1": 3.0, "m2": 4.0},
            }
        )

        # Confirm the expected state after attaching data
        data = (
            self.batch.lookup_data()
            .df.set_index(["arm_name", "metric_name"])
            .to_dict(orient="index")
        )

        self.assertEqual(1.0, data[(arm1_name, "m1")]["mean"])
        self.assertEqual(2.0, data[(arm1_name, "m2")]["mean"])
        self.assertEqual(3.0, data[(arm2_name, "m1")]["mean"])
        self.assertEqual(4.0, data[(arm2_name, "m2")]["mean"])



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from unittest import mock

import numpy as np

import torch

from ax.core.base_trial import TrialStatus
from ax.core.observation import (
    ObservationData,
    ObservationFeatures,
    recombine_observations,
)
from ax.modelbridge.map_torch import MapTorchModelBridge
from ax.models.torch_base import TorchGenResults, TorchModel
from ax.utils.common.constants import Keys
from ax.utils.common.testutils import TestCase
from ax.utils.testing.core_stubs import (
    get_branin_arms,
    get_branin_experiment_with_timestamp_map_metric,
)


class MapTorchModelBridgeTest(TestCase):
    def test_TorchModelBridge(self) -> None:
        experiment = get_branin_experiment_with_timestamp_map_metric(rate=0.5)
        for i in range(3):
            trial = experiment.new_trial().add_arm(arm=get_branin_arms(n=1, seed=i)[0])
            trial.run()

        for _ in range(2):
            # each time we call fetch, we grab another timestamp
            experiment.fetch_data()

        for i in range(3):
            experiment.trials[i].mark_as(status=TrialStatus.COMPLETED)

        experiment.attach_data(data=experiment.fetch_data())
        modelbridge = MapTorchModelBridge(
            experiment=experiment,
            search_space=experiment.search_space,
            data=experiment.lookup_data(),
            model=TorchModel(),
            transforms=[],
            fit_out_of_design=True,
            default_model_gen_options={"target_map_values": {"timestamp": 4.0}},
        )
        # Check map data is converted to observations, that we get one Observation
        # per row of MapData
        # pyre-fixme[16]: `Data` has no attribute `map_df`.
        map_df = experiment.lookup_data().map_df
        objective_df = map_df[map_df["metric_name"] == "branin_map"]
        self.assertEqual(len(modelbridge.get_training_data()), len(objective_df))

        # Test _gen
        model = mock.MagicMock(TorchModel, autospec=True, instance=True)
        model.gen.return_value = TorchGenResults(
            points=torch.tensor([[0.0, 0.0]]),
            weights=torch.tensor([1.0]),
            gen_metadata={},
        )
        model.predict.return_value = (
            torch.tensor([[0.0, 0.0]]),
            torch.tensor([[[1.0, 0.0], [0.0, 1.0]]]),
        )
        modelbridge.model = model
        gen_results = modelbridge._gen(
            n=1,
            search_space=experiment.search_space,
            optimization_config=experiment.optimization_config,
            pending_observations={},
            fixed_features=ObservationFeatures({}),
        )
        gen_args = model.gen.mock_calls[0][2]
        self.assertEqual(
            gen_args["torch_opt_config"].model_gen_options[Keys.ACQF_KWARGS],
            {"map_dim_to_target": {2: 4.0}},
        )
        self.assertEqual(
            gen_results.observation_features[0].parameters, {"x1": 0.0, "x2": 0.0}
        )

        # Test _predict
        model = mock.MagicMock(TorchModel, autospec=True, instance=True)
        model.predict.return_value = (
            torch.tensor([[0.0, 0.0]]),
            torch.tensor([[[1.0, 0.0], [0.0, 1.0]]]),
        )
        modelbridge.model = model
        features = [
            ObservationFeatures(parameters={"x1": 0.0, "x2": 0.0}),
            ObservationFeatures(parameters={"x1": 1.0, "x2": 1.0}),
        ]
        modelbridge._predict(features)
        predict_args = model.predict.mock_calls[0][2]
        # check that the target map value is inserted before prediction
        self.assertTrue(
            torch.allclose(
                predict_args["X"],
                torch.tensor([[0.0, 0.0, 4.0], [1.0, 1.0, 4.0]]).to(predict_args["X"]),
            )
        )

        # Test _cross_validate
        features = [
            ObservationFeatures(parameters={"x1": 0.0, "x2": 0.0, "timestamp": 0.0}),
            ObservationFeatures(parameters={"x1": 1.0, "x2": 1.0, "timestamp": 2.0}),
        ]
        data = [
            ObservationData(
                metric_names=["branin_map"],
                means=np.array([0.0]),
                covariance=np.array([[1.0]]),
            ),
            ObservationData(
                metric_names=["branin_map"],
                means=np.array([2.0]),
                covariance=np.array([[1.0]]),
            ),
        ]
        test_features = [
            ObservationFeatures(parameters={"x1": 0.0, "x2": 0.0, "timestamp": 1.0}),
            ObservationFeatures(parameters={"x1": 1.0, "x2": 1.0, "timestamp": 3.0}),
        ]
        test_data = [
            ObservationData(
                metric_names=["branin_map"],
                means=np.array([1.0]),
                covariance=np.array([[1.0]]),
            ),
            ObservationData(
                metric_names=["branin_map", "branin"],
                means=np.array([3.0, 2.0]),
                covariance=np.array([[1.0, 0.0], [0.0, 1.0]]),
            ),
        ]
        cv_training_data = recombine_observations(features, data)
        with mock.patch(
            "ax.modelbridge.torch.TorchModelBridge._cross_validate",
            return_value=test_data,
        ):
            cv_obs_data = modelbridge._cross_validate(
                search_space=experiment.search_space,
                cv_training_data=cv_training_data,
                cv_test_points=test_features,
            )
            # check that the out-of-design metric is deleted
            self.assertEqual(cv_obs_data[1].metric_names, ["branin"])



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import argparse
import os
import pkgutil
import re
from typing import Set


# Paths are relative to top-level Ax directory (which is passed into fxn below)
SPHINX_RST_PATH = os.path.join("github", "sphinx", "source")

AX_LIBRARY_PATH = "ax"

# Different because of ShipIt config
OSS_SPHINX_RST_PATH = os.path.join("sphinx", "source")


# Regex for automodule directive used in Sphinx docs
AUTOMODULE_REGEX = re.compile(r"\.\. automodule:: ([\.\w]*)")

# Modules to exclude from validation
EXCLUDE_MODULES = {
    "ax.utils.testing.manifest",
}


def parse_rst(rst_filename: str) -> Set[str]:
    """Extract automodule directives from rst."""
    ret = set()
    with open(rst_filename, "r") as f:
        lines = f.readlines()
        for line in lines:
            line = line.strip()
            name = AUTOMODULE_REGEX.findall(line)
            if name:
                ret.add(name[0])
    return ret


def validate_complete_sphinx(path_to_ax: str) -> None:
    """Validate that Sphinx-based API documentation is complete.

    * Every top-level module (e.g., core, models, etc.) should have corresponding rst
      file.
    * Every single non-package (i.e. py file) module should be included in rst file with
      `automodule::` directive. Sphinx will then automatically include all members from
       the module in the documentation.

    Note: this function does not validate any documentation for the 'ax' module.

    Args:
      path_to_ax: the path to the top-level ax directory (directory that includes ax
        library, sphinx, website, etc.).

    """
    # Load top-level modules used in Ax (e.g., core, models; exclude 'fb' and 'version')
    modules = {
        modname
        for importer, modname, ispkg in pkgutil.walk_packages(
            path=[AX_LIBRARY_PATH], onerror=lambda x: None
        )
        if modname not in {"fb", "version", "__version__"}
    }

    # Load all rst files (these contain the documentation for Sphinx)
    try:
        rstpath = os.path.join(path_to_ax, SPHINX_RST_PATH)
        rsts = {
            f.replace(".rst", "") for f in os.listdir(rstpath) if f.endswith(".rst")
        }
    except FileNotFoundError:
        rstpath = os.path.join(path_to_ax, OSS_SPHINX_RST_PATH)
        rsts = {
            f.replace(".rst", "") for f in os.listdir(rstpath) if f.endswith(".rst")
        }

    # Verify that all top-level modules have a corresponding rst
    modules_without_rsts = modules.difference(rsts)
    assert (
        len(modules_without_rsts) == 0
    ), f"Not all modules have corresponding rst: {modules_without_rsts}"

    # Track all modules that are not in docs (so can print all)
    modules_not_in_docs = []

    # Iterate over top-level modules
    for module in modules.intersection(rsts):
        # Parse rst & extract all modules use automodule directive
        modules_in_rst = parse_rst(os.path.join(rstpath, module + ".rst"))

        # Extract all non-package modules
        for _importer, modname, ispkg in pkgutil.walk_packages(
            path=[os.path.join(AX_LIBRARY_PATH, module)],  # ax.__path__[0], module),
            prefix="ax." + module + ".",
            onerror=lambda x: None,
        ):
            if (
                not ispkg
                and ".tests" not in modname
                and modname not in modules_in_rst.union(EXCLUDE_MODULES)
            ):
                modules_not_in_docs.append(modname)

    assert len(modules_not_in_docs) == 0, "Not all modules are documented: {}".format(
        modules_not_in_docs
    )


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Validate that Sphinx documentation is complete."
    )
    parser.add_argument(
        "-p",
        "--path",
        metavar="path",
        required=True,
        help="Path to the top-level ax directory.",
    )
    args = parser.parse_args()
    validate_complete_sphinx(args.path)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from ax.core.parameter import ChoiceParameter, ParameterType, RangeParameter
from ax.core.search_space import SearchSpace
from ax.exceptions.core import UnsupportedError
from ax.modelbridge.transforms.cap_parameter import CapParameter
from ax.utils.common.testutils import TestCase
from ax.utils.testing.core_stubs import get_robust_search_space


class CapParameterTest(TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.search_space = SearchSpace(
            parameters=[
                RangeParameter(
                    "a", lower=1, upper=3, parameter_type=ParameterType.FLOAT
                ),
                ChoiceParameter(
                    "b", parameter_type=ParameterType.STRING, values=["a", "b", "c"]
                ),
            ]
        )

    def test_transform_search_space(self) -> None:
        t = CapParameter(
            search_space=self.search_space,
            observations=[],
            config={"a": "2"},
        )
        t.transform_search_space(self.search_space)
        # pyre-fixme[16]: Optional type has no attribute `upper`.
        self.assertEqual(self.search_space.parameters.get("a").upper, 2)
        t2 = CapParameter(
            search_space=self.search_space,
            observations=[],
            config={"b": "2"},
        )
        with self.assertRaises(NotImplementedError):
            t2.transform_search_space(self.search_space)

    def test_w_parameter_distributions(self) -> None:
        rss = get_robust_search_space()
        # Transform a non-distributional parameter.
        t = CapParameter(
            search_space=rss,
            observations=[],
            config={"z": "2"},
        )
        t.transform_search_space(rss)
        # pyre-fixme[16]: Optional type has no attribute `upper`.
        self.assertEqual(rss.parameters.get("z").upper, 2)
        # Error with distributional parameter.
        t = CapParameter(
            search_space=rss,
            observations=[],
            config={"x": "2"},
        )
        with self.assertRaisesRegex(UnsupportedError, "transform is not supported"):
            t.transform_search_space(rss)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import os

from setuptools import find_packages, setup

PINNED_BOTORCH_VERSION = "0.11.3"

if os.environ.get("ALLOW_BOTORCH_LATEST"):
    # allows a more recent previously installed version of botorch to remain
    # if there is no previously installed version, installs the latest release
    botorch_req = f"botorch>={PINNED_BOTORCH_VERSION}"
else:
    botorch_req = f"botorch=={PINNED_BOTORCH_VERSION}"


REQUIRES = [
    botorch_req,
    "jinja2",  # also a Plotly dep
    "pandas",
    "scipy",
    "scikit-learn",
    "ipywidgets",
    # Needed for compatibility with ipywidgets >= 8.0.0
    "plotly>=5.12.0",
    "typeguard",
    "pyre-extensions",
]

# pytest-cov requires pytest >= 3.6
DEV_REQUIRES = [
    "beautifulsoup4",
    "black==24.2.0",
    "flake8",
    "hypothesis",
    "Jinja2",
    "pyfakefs",
    "pytest>=4.6",
    "pytest-cov",
    "sphinx==5.3.0",
    "sphinx-autodoc-typehints==1.19.5",
    "torchvision>=0.5.0",
    "nbconvert",
    "jupyter-client==6.1.12",
]

MYSQL_REQUIRES = ["SQLAlchemy==1.4.17"]

NOTEBOOK_REQUIRES = ["jupyter"]

UNITTEST_MINIMAL_REQUIRES = [
    "tensorboard",  # For tensorboard unit tests.
    "torchvision",  # For torchvision unit tests.
    "torchx",  # For torchx unit tests.
]

UNITTEST_REQUIRES = (
    DEV_REQUIRES + MYSQL_REQUIRES + NOTEBOOK_REQUIRES + UNITTEST_MINIMAL_REQUIRES
)

TUTORIAL_REQUIRES = UNITTEST_REQUIRES + [
    "ray",  # Required for building RayTune tutorial notebook.
    "tabulate",  # Required for building RayTune tutorial notebook.
    "pyarrow",  # Required for building RayTune tutorial notebook.
    "tensorboardX",  # Required for building RayTune tutorial notebook.
    "matplotlib",  # Required for building Multi-objective tutorial notebook.
    "pyro-ppl",  # Required for to call run_inference.
    "pytorch-lightning",  # For the early stopping tutorial.
    "papermill",  # For executing the tutorials.
    "submitit",  # Required for building the SubmitIt notebook.
]


def local_version(version):
    """
    Patch in a version that can be uploaded to test PyPI
    """
    return ""


def setup_package() -> None:
    """Used for installing the Ax package."""

    with open("README.md", "r") as fh:
        long_description = fh.read()

    setup(
        name="ax-platform",
        description="Adaptive Experimentation",
        author="Facebook, Inc.",
        license="MIT",
        url="https://github.com/facebook/Ax",
        keywords=["Experimentation", "Optimization"],
        classifiers=[
            "Development Status :: 4 - Beta",
            "Operating System :: POSIX :: Linux",
            "Operating System :: MacOS :: MacOS X",
            "Programming Language :: Python :: 3",
        ],
        long_description=long_description,
        long_description_content_type="text/markdown",
        python_requires=">=3.10",
        install_requires=REQUIRES,
        packages=find_packages(),
        package_data={
            # include all js, css, and html files in the package
            "": ["*.js", "*.css", "*.html"]
        },
        extras_require={
            "dev": DEV_REQUIRES,
            "mysql": MYSQL_REQUIRES,
            "notebook": NOTEBOOK_REQUIRES,
            "unittest": UNITTEST_REQUIRES,
            "unittest_minimal": UNITTEST_MINIMAL_REQUIRES,
            "tutorial": TUTORIAL_REQUIRES,
        },
        use_scm_version={
            "write_to": "ax/version.py",
            "local_scheme": local_version,
        },
        setup_requires=["setuptools_scm"],
    )


if __name__ == "__main__":
    setup_package()



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from typing import Any

import numpy as np
import torch
from torch import Tensor


# pyre-fixme[2]: Parameter annotation cannot be `Any`.
def generic_equals(first: Any, second: Any) -> bool:
    if isinstance(first, Tensor):
        return isinstance(second, Tensor) and torch.equal(first, second)
    if isinstance(first, np.ndarray):
        return isinstance(second, np.ndarray) and np.array_equal(
            first, second, equal_nan=True
        )
    if isinstance(first, dict):
        return isinstance(second, dict) and generic_equals(
            sorted(first.items()), sorted(second.items())
        )
    if isinstance(first, (tuple, list)):
        if type(first) is not type(second) or len(first) != len(second):
            return False
        for f, s in zip(first, second):
            if not generic_equals(f, s):
                return False
        return True
    return first == second



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from unittest import mock

import numpy as np
from ax.core.observation import ObservationFeatures
from ax.core.types import TEvaluationOutcome, TParameterization
from ax.modelbridge.prediction_utils import predict_at_point, predict_by_features
from ax.service.ax_client import AxClient
from ax.service.utils.instantiation import ObjectiveProperties
from ax.utils.common.testutils import TestCase
from ax.utils.common.typeutils import not_none


class TestPredictionUtils(TestCase):
    """Tests prediction utilities."""

    def test_predict_at_point(self) -> None:
        ax_client = _set_up_client_for_get_model_predictions_no_next_trial()
        _attach_completed_trials(ax_client)
        ax_client.get_model_predictions()  # Ensures model is instantiated

        observation_features = ObservationFeatures(parameters={"x1": 0.3, "x2": 0.5})
        y_hat, se_hat = predict_at_point(
            model=not_none(ax_client.generation_strategy.model),
            obsf=observation_features,
            metric_names={"test_metric1"},
        )

        self.assertEqual(len(y_hat), 1)
        self.assertEqual(len(se_hat), 1)

        y_hat, se_hat = predict_at_point(
            model=not_none(ax_client.generation_strategy.model),
            obsf=observation_features,
            metric_names={"test_metric1", "test_metric2", "test_metric:agg"},
            scalarized_metric_config=[
                {
                    "name": "test_metric:agg",
                    "weight": {"test_metric1": 0.5, "test_metric2": 0.5},
                },
            ],
        )
        self.assertEqual(len(y_hat), 3)
        self.assertEqual(len(se_hat), 3)

        y_hat, se_hat = predict_at_point(
            model=not_none(ax_client.generation_strategy.model),
            obsf=observation_features,
            metric_names={"test_metric1"},
            scalarized_metric_config=[
                {
                    "name": "test_metric:agg",
                    "weight": {"test_metric1": 0.5, "test_metric2": 0.5},
                },
            ],
        )
        self.assertEqual(len(y_hat), 1)
        self.assertEqual(len(se_hat), 1)

    def test_predict_by_features(self) -> None:
        ax_client = _set_up_client_for_get_model_predictions_no_next_trial()
        _attach_completed_trials(ax_client)
        ax_client.get_model_predictions()  # Ensures model is instantiated

        observation_features_dict = {
            18: ObservationFeatures(parameters={"x1": 0.3, "x2": 0.5}),
            19: ObservationFeatures(parameters={"x1": 0.4, "x2": 0.5}),
            20: ObservationFeatures(parameters={"x1": 0.8, "x2": 0.5}),
        }
        predictions_map = predict_by_features(
            model=not_none(ax_client.generation_strategy.model),
            label_to_feature_dict=observation_features_dict,
            metric_names={"test_metric1"},
        )
        self.assertEqual(len(predictions_map), 3)

    @mock.patch("ax.modelbridge.random.RandomModelBridge.predict")
    @mock.patch("ax.modelbridge.random.RandomModelBridge")
    # pyre-fixme[3]: Return type must be annotated.
    def test_predict_by_features_with_non_predicting_model(
        self,
        # pyre-fixme[2]: Parameter must be annotated.
        model_bridge_mock,
        # pyre-fixme[2]: Parameter must be annotated.
        predict_mock,
    ):
        ax_client = _set_up_client_for_get_model_predictions_no_next_trial()
        _attach_completed_trials(ax_client)

        # Do not call get_next_trial or get_model_predictions.
        # This test is for handling the use case where no model
        # is instantiated.

        observation_features_dict = {
            18: ObservationFeatures(parameters={"x1": 0.3, "x2": 0.5}),
            19: ObservationFeatures(parameters={"x1": 0.4, "x2": 0.5}),
            20: ObservationFeatures(parameters={"x1": 0.8, "x2": 0.5}),
        }

        predict_mock.side_effect = NotImplementedError()
        self.assertRaises(
            NotImplementedError,
            predict_by_features,
            **{
                "model": model_bridge_mock,
                "label_to_feature_dict": observation_features_dict,
                "metric_names": ["test_metric1"],
            },
        )


# Utility functions for testing get_model_predictions without calling
# get_next_trial. Create Ax Client with an experiment where
# num_initial_trials kwarg is zero. Note that this kwarg is
# needed to be able to instantiate the model for the first time
# without calling get_next_trial().
def _set_up_client_for_get_model_predictions_no_next_trial() -> AxClient:
    ax_client = AxClient()
    ax_client.create_experiment(
        name="test_experiment",
        choose_generation_strategy_kwargs={"num_initialization_trials": 0},
        parameters=[
            {
                "name": "x1",
                "type": "range",
                "bounds": [0.0, 1.0],
            },
            {
                "name": "x2",
                "type": "range",
                "bounds": [0.1, 1.0],
            },
        ],
        objectives={"test_metric1": ObjectiveProperties(minimize=False)},
        outcome_constraints=["test_metric2 <= 1.5"],
    )

    return ax_client


def _attach_completed_trials(ax_client: AxClient) -> None:
    # Attach completed trials
    trial1: TParameterization = {"x1": 0.1, "x2": 0.1}
    parameters, trial_index = ax_client.attach_trial(trial1)
    ax_client.complete_trial(
        trial_index=trial_index, raw_data=_evaluate_test_metrics(parameters)
    )

    trial2: TParameterization = {"x1": 0.2, "x2": 0.1}
    parameters, trial_index = ax_client.attach_trial(trial2)
    ax_client.complete_trial(
        trial_index=trial_index, raw_data=_evaluate_test_metrics(parameters)
    )


# Test metric evaluation method
def _evaluate_test_metrics(parameters: TParameterization) -> TEvaluationOutcome:
    x = np.array([parameters.get(f"x{i+1}") for i in range(2)])
    return {"test_metric1": (x[0] / x[1], 0.0), "test_metric2": (x[0] + x[1], 0.0)}



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import pandas as pd
import plotly.graph_objects as go

from ax.analysis.analysis_report import AnalysisReport

from ax.analysis.base_analysis import BaseAnalysis
from ax.analysis.base_plotly_visualization import BasePlotlyVisualization

from ax.modelbridge.registry import Models
from ax.utils.common.testutils import TestCase

from ax.utils.common.timeutils import current_timestamp_in_millis
from ax.utils.testing.core_stubs import get_branin_experiment
from ax.utils.testing.mock import fast_botorch_optimize


class TestCrossValidationPlot(TestCase):

    class TestAnalysis(BaseAnalysis):
        def get_df(self) -> pd.DataFrame:
            return pd.DataFrame()

    class TestPlotlyVisualization(BasePlotlyVisualization):
        def get_df(self) -> pd.DataFrame:
            return pd.DataFrame()

        def get_fig(self) -> go.Figure:
            return go.Figure()

    @fast_botorch_optimize
    def setUp(self) -> None:
        super().setUp()

        self.exp = get_branin_experiment(with_batch=True)
        self.exp.trials[0].run()
        self.model = Models.BOTORCH_MODULAR(
            # Model bridge kwargs
            experiment=self.exp,
            data=self.exp.fetch_data(),
        )

        self.test_analysis = self.TestAnalysis(experiment=self.exp)
        self.test_plotly_visualization = self.TestPlotlyVisualization(
            experiment=self.exp
        )

    def test_init_analysis_report(self) -> None:
        analysis_report = AnalysisReport(
            experiment=self.exp,
            analyses=[self.test_analysis, self.test_plotly_visualization],
        )

        self.assertEqual(len(analysis_report.analyses), 2)

        self.assertIsInstance(analysis_report.analyses[0], BaseAnalysis)
        self.assertIsInstance(analysis_report.analyses[1], BasePlotlyVisualization)

        self.assertIsNone(analysis_report.time_started)
        self.assertIsNone(analysis_report.time_completed)
        self.assertFalse(analysis_report.report_completed)

    def test_execute_analysis_report(self) -> None:
        analysis_report = AnalysisReport(
            experiment=self.exp,
            analyses=[self.test_analysis, self.test_plotly_visualization],
        )

        self.assertEqual(len(analysis_report.analyses), 2)

        results = analysis_report.run_analysis_report()
        self.assertEqual(len(results), 2)

        self.assertIsNotNone(analysis_report.time_started)
        self.assertIsNotNone(analysis_report.time_completed)
        self.assertTrue(analysis_report.report_completed)

        # assert no plot is returned for BaseAnalysis
        self.assertIsInstance(results[0][1], pd.DataFrame)
        self.assertIsNone(results[0][2])

        # assert plot is returned for BasePlotlyVisualization
        self.assertIsInstance(results[1][1], pd.DataFrame)
        self.assertIsInstance(results[1][2], go.Figure)

    def test_analysis_report_repeated_execute(self) -> None:
        analysis_report = AnalysisReport(
            experiment=self.exp,
            analyses=[self.test_analysis, self.test_plotly_visualization],
        )

        self.assertEqual(len(analysis_report.analyses), 2)

        _ = analysis_report.run_analysis_report()

        saved_start = analysis_report.time_started
        self.assertIsNotNone(saved_start)

        # ensure analyses are not re-ran
        _ = analysis_report.run_analysis_report()
        self.assertEqual(saved_start, analysis_report.time_started)

    def test_no_analysis_report(self) -> None:
        analysis_report = AnalysisReport(
            experiment=self.exp,
            analyses=[],
        )

        self.assertEqual(len(analysis_report.analyses), 0)

        results = analysis_report.run_analysis_report()
        self.assertEqual(len(results), 0)

    def test_singleton_analysis_report(self) -> None:
        analysis_report = AnalysisReport(
            experiment=self.exp,
            analyses=[self.test_plotly_visualization],
        )

        self.assertEqual(len(analysis_report.analyses), 1)

        results = analysis_report.run_analysis_report()
        self.assertEqual(len(results), 1)

        # assert plot is returned for BasePlotlyVisualization
        self.assertIsInstance(results[0][1], pd.DataFrame)
        self.assertIsInstance(results[0][2], go.Figure)

    def test_create_report_as_completed(self) -> None:
        analysis_report = AnalysisReport(
            experiment=self.exp,
            analyses=[self.test_analysis, self.test_plotly_visualization],
            time_started=current_timestamp_in_millis(),
            time_completed=current_timestamp_in_millis(),
        )

        self.assertIsNotNone(analysis_report.time_started)
        self.assertIsNotNone(analysis_report.time_completed)
        self.assertTrue(analysis_report.report_completed)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from contextlib import contextmanager
from typing import Any, Callable, TypeVar

from unittest.mock import MagicMock, patch


T = TypeVar("T")
C = TypeVar("C")


@contextmanager
def mock_patch_method_original(
    mock_path: str,
    original_method: Callable[..., T],
) -> MagicMock:
    """Context manager for patching a method returning type T on class C,
    to track calls to it while still executing the original method. There
    is not a native way to do this with `mock.patch`.
    """

    def side_effect(self: C, *args: Any, **kwargs: Any) -> T:
        # pyre-ignore[16]: Anonymous callable has no attribute `self`
        # (We can ignore because we expect C to be a class).
        side_effect.self = self
        return original_method(self, *args, **kwargs)

    patcher = patch(mock_path, autospec=True, side_effect=side_effect)
    yield patcher.start()
    patcher.stop()
